from __future__ import annotations

import json
import tempfile
import tomllib
import unittest
from pathlib import Path
from unittest.mock import patch

from llm_launchpad.protocol.enums import BackendType, OperationType
from llm_launchpad.protocol.events import LogEvent, OperationCompleteEvent
from llm_launchpad.protocol.models import DeploymentConfig
from llm_launchpad.protocol.models import EndpointInfo
from llm_launchpad.protocol.models import StorageSnapshot, StoredModelInfo
from llm_launchpad.tui.app import TuiApp, _deploy_connection_summary_lines


class TuiAppStorageCacheTests(unittest.TestCase):
    def test_packaging_config_includes_tui_theme_css(self) -> None:
        pyproject = Path(__file__).resolve().parents[1] / "pyproject.toml"
        payload = tomllib.loads(pyproject.read_text(encoding="utf-8"))

        package_data = payload["tool"]["setuptools"]["package-data"]
        self.assertEqual(package_data["llm_launchpad.tui"], ["theme.tcss"])

    def test_on_mount_launches_main_menu_without_authenticated_preflight(self) -> None:
        app = TuiApp()
        app._version = "1.2.3"
        app._orchestrator = type(
            "FakeOrchestrator",
            (),
            {
                "preflight": staticmethod(lambda: (False, "", "Modal authentication missing. Run: modal setup")),
            },
        )()

        with (
            patch("llm_launchpad.tui.app.ModalBackend.is_cli_available", return_value=True),
            patch("llm_launchpad.tui.app.ModalBackend.get_username", return_value="default"),
            patch.object(app, "push_screen", return_value=None) as push_screen,
            patch.object(app, "notify", return_value=None) as notify,
            patch.object(app, "exit", return_value=None) as exit_mock,
        ):
            app.on_mount()

        notify.assert_not_called()
        exit_mock.assert_not_called()
        push_screen.assert_called_once()

    def test_deploy_connection_summary_lines_for_llamacpp_are_simple_and_boxed(self) -> None:
        config = DeploymentConfig(
            backend=BackendType.LLAMACPP,
            repo_id="Edge-Quant/Nanbeige4.1-3B-Q4_K_M-GGUF",
            quant="Q4_K_M",
            app_name="llamacpp-edge-quant-nanbeige4-1-3b-q4-k-m-gguf",
            instance_name="edge-quant-nanbeige4-1-3b-q4-k-m-gguf",
        )
        lines = _deploy_connection_summary_lines(
            config,
            "https://example.modal.run",
        )
        joined = "\n".join(lines)
        self.assertIn("=== OpenAI-compatible ===", joined)
        self.assertIn("Base URL: https://example.modal.run/v1", joined)
        self.assertIn("Model ID: Nanbeige4.1-3B-Q4_K_M-GGUF", joined)
        self.assertIn("Display name: Nanbeige4.1-3B-Q4_K_M-GGUF (Q4_K_M)", joined)
        self.assertIn("API key: (leave blank; no auth by default)", joined)
        self.assertIn("=========================", joined)
        self.assertNotIn("OpenCode custom provider:", joined)

    def test_deploy_connection_summary_lines_for_vllm_use_served_model_alias(self) -> None:
        config = DeploymentConfig(
            backend=BackendType.VLLM,
            model_name="Nanbeige/Nanbeige4.1-3B",
            served_model_name="Nanbeige4.1-3B",
            app_name="vllm-nanbeige4-1-3b",
            instance_name="nanbeige4-1-3b",
        )
        lines = _deploy_connection_summary_lines(
            config,
            "https://example.modal.run",
        )
        joined = "\n".join(lines)
        self.assertIn("Base URL: https://example.modal.run/v1", joined)
        self.assertIn("Model ID: Nanbeige4.1-3B", joined)
        self.assertIn("Display name: Nanbeige4.1-3B", joined)

    def test_run_deploy_warmup_uses_function_slug_for_default_url(self) -> None:
        warmup_urls: list[str] = []
        app = TuiApp()
        app._username = "alice"
        app._orchestrator = type(
            "FakeOrchestrator",
            (),
            {
                "deploy": staticmethod(
                    lambda _config: [
                        OperationCompleteEvent(success=True, operation=OperationType.DEPLOY)
                    ]
                ),
                "warmup": staticmethod(
                    lambda backend, server_url, *_args, **_kwargs: warmup_urls.append(server_url) or []
                ),
            },
        )()

        config = DeploymentConfig(
            backend=BackendType.LLAMACPP,
            do_deploy=True,
            do_warmup=True,
            app_name="llamacpp-test",
            function_slug="alpha-bravo",
        )
        with patch("llm_launchpad.tui.app._dispatch_event", return_value=None):
            app._run_deploy(config, monitor=object())

        self.assertEqual(
            warmup_urls,
            ["https://alice--llamacpp-test-serve-alpha-bravo.modal.run"],
        )

    def test_run_deploy_warmup_prefers_created_web_function_url_over_printed_endpoint(self) -> None:
        warmup_urls: list[str] = []
        app = TuiApp()
        app._username = "alice"
        created_dev_url = "https://alice--llamacpp-test-serve-alpha-bravo-dev.modal.run"
        created_prod_url = "https://alice--llamacpp-test-serve-abcd1234.modal.run"
        guessed_url = "https://alice--llamacpp-test-serve-alpha-bravo.modal.run"
        app._orchestrator = type(
            "FakeOrchestrator",
            (),
            {
                "deploy": staticmethod(
                    lambda _config: [
                        LogEvent(line=f"└── 🔨 Created web function serve => {created_dev_url}"),
                        LogEvent(
                            line=(
                                "└── 🔨 Created web function serve => "
                                f"{created_prod_url} (label truncated)"
                            )
                        ),
                        LogEvent(line=f"   Endpoint base URL: {guessed_url}"),
                        OperationCompleteEvent(success=True, operation=OperationType.DEPLOY),
                    ]
                ),
                "warmup": staticmethod(
                    lambda backend, server_url, *_args, **_kwargs: warmup_urls.append(server_url) or []
                ),
            },
        )()

        config = DeploymentConfig(
            backend=BackendType.LLAMACPP,
            do_deploy=True,
            do_warmup=True,
            app_name="llamacpp-test",
            function_slug="alpha-bravo",
        )
        with patch("llm_launchpad.tui.app._dispatch_event", return_value=None):
            app._run_deploy(config, monitor=object())

        self.assertEqual(warmup_urls, [created_prod_url])

    def test_run_deploy_suppresses_intermediate_deploy_completion_when_warmup_follows(self) -> None:
        app = TuiApp()
        app._username = "alice"
        app._orchestrator = type(
            "FakeOrchestrator",
            (),
            {
                "deploy": staticmethod(
                    lambda _config: [
                        OperationCompleteEvent(success=True, operation=OperationType.DEPLOY)
                    ]
                ),
                "warmup": staticmethod(
                    lambda *_args, **_kwargs: [
                        LogEvent(line="Probing readiness at: https://example.com/v1/completions"),
                        OperationCompleteEvent(success=True, operation=OperationType.WARMUP),
                    ]
                ),
            },
        )()

        config = DeploymentConfig(
            backend=BackendType.LLAMACPP,
            do_deploy=True,
            do_warmup=True,
            app_name="llamacpp-test",
            function_slug="alpha-bravo",
        )
        with patch("llm_launchpad.tui.app._dispatch_event", return_value=None) as dispatch:
            app._run_deploy(config, monitor=object())

        dispatched_events = [call.args[1] for call in dispatch.call_args_list]
        self.assertFalse(
            any(
                isinstance(event, OperationCompleteEvent)
                and event.operation == OperationType.DEPLOY
                and event.success
                for event in dispatched_events
            )
        )
        self.assertTrue(
            any(
                isinstance(event, OperationCompleteEvent)
                and event.operation == OperationType.WARMUP
                and event.success
                for event in dispatched_events
            )
        )

    def test_run_deploy_emits_connection_summary_on_warmup_success(self) -> None:
        app = TuiApp()
        app._username = "alice"
        app._orchestrator = type(
            "FakeOrchestrator",
            (),
            {
                "deploy": staticmethod(
                    lambda _config: [
                        LogEvent(line="└── 🔨 Created web function serve => https://alice--llamacpp-test-serve-abcd.modal.run"),
                        OperationCompleteEvent(success=True, operation=OperationType.DEPLOY),
                    ]
                ),
                "warmup": staticmethod(
                    lambda *_args, **_kwargs: [
                        OperationCompleteEvent(
                            success=True,
                            operation=OperationType.WARMUP,
                            data={"url": "https://alice--llamacpp-test-serve-abcd.modal.run"},
                        )
                    ]
                ),
            },
        )()

        config = DeploymentConfig(
            backend=BackendType.LLAMACPP,
            repo_id="Edge-Quant/Nanbeige4.1-3B-Q4_K_M-GGUF",
            quant="Q4_K_M",
            do_deploy=True,
            do_warmup=True,
            app_name="llamacpp-test",
            instance_name="llamacpp-test",
            function_slug="alpha-bravo",
        )
        with patch("llm_launchpad.tui.app._dispatch_event", return_value=None) as dispatch:
            app._run_deploy(config, monitor=object())

        dispatched_events = [call.args[1] for call in dispatch.call_args_list]
        summary_lines = [
            event.line
            for event in dispatched_events
            if isinstance(event, LogEvent)
            and (
                "=== OpenAI-compatible ===" in event.line
                or "Base URL:" in event.line
                or "Model ID:" in event.line
                or "Display name:" in event.line
            )
        ]
        self.assertTrue(any("=== OpenAI-compatible ===" in line for line in summary_lines))
        self.assertTrue(any("Model ID: Nanbeige4.1-3B-Q4_K_M-GGUF" in line for line in summary_lines))
        self.assertTrue(any("Display name: Nanbeige4.1-3B-Q4_K_M-GGUF (Q4_K_M)" in line for line in summary_lines))

    def test_run_deploy_syncs_opencode_on_warmup_success(self) -> None:
        app = TuiApp()
        app._username = "alice"
        app._orchestrator = type(
            "FakeOrchestrator",
            (),
            {
                "deploy": staticmethod(
                    lambda _config: [
                        OperationCompleteEvent(success=True, operation=OperationType.DEPLOY),
                    ]
                ),
                "warmup": staticmethod(
                    lambda *_args, **_kwargs: [
                        OperationCompleteEvent(
                            success=True,
                            operation=OperationType.WARMUP,
                            data={"url": "https://alice--llamacpp-test-serve-abcd.modal.run"},
                        )
                    ]
                ),
            },
        )()

        config = DeploymentConfig(
            backend=BackendType.LLAMACPP,
            repo_id="Edge-Quant/Nanbeige4.1-3B-Q4_K_M-GGUF",
            quant="Q4_K_M",
            do_deploy=True,
            do_warmup=True,
            app_name="llamacpp-test",
            instance_name="llamacpp-test",
            function_slug="alpha-bravo",
        )
        with (
            patch("llm_launchpad.tui.app._dispatch_event", return_value=None),
            patch.object(app, "_load_visible_launchpad_rows", return_value=[]),
            patch.object(app, "_sync_opencode") as sync_mock,
        ):
            app._run_deploy(config, monitor=object())

        sync_mock.assert_called_once()
        self.assertEqual(sync_mock.call_args.kwargs["target_app_name"], "llamacpp-test")
        self.assertEqual(
            sync_mock.call_args.kwargs["target_url"],
            "https://alice--llamacpp-test-serve-abcd.modal.run",
        )

    def test_list_instances_merges_cached_deploy_connection_summary(self) -> None:
        with tempfile.TemporaryDirectory() as tmp:
            app = TuiApp()
            app._deploy_connection_cache_path = Path(tmp) / "deployment_connection_summaries.json"
            app._deploy_connection_cache = {}
            config = DeploymentConfig(
                backend=BackendType.LLAMACPP,
                repo_id="Edge-Quant/Nanbeige4.1-3B-Q4_K_M-GGUF",
                quant="Q4_K_M",
                app_name="llamacpp-edge-quant",
                instance_name="edge-quant",
                served_model_name="Nanbeige4.1-3B-Q4_K_M-GGUF",
            )
            app._cache_deploy_connection_summary(config, "https://alice--llamacpp-edge-quant-serve.modal.run")

            rows = [
                EndpointInfo(
                    name="llamacpp-edge-quant",
                    app_id="ap-123",
                    state="running",
                    backend=BackendType.LLAMACPP,
                    instance_name="edge-quant",
                )
            ]
            with patch("llm_launchpad.tui.app.ModalBackend.list_apps", return_value=rows):
                merged = app.list_instances()

            self.assertEqual(len(merged), 1)
            self.assertEqual(
                merged[0].web_url,
                "https://alice--llamacpp-edge-quant-serve.modal.run/v1",
            )
            self.assertEqual(merged[0].served_model_name, "Nanbeige4.1-3B-Q4_K_M-GGUF")
            self.assertEqual(
                merged[0].display_name,
                "Nanbeige4.1-3B-Q4_K_M-GGUF (Q4_K_M)",
            )

    def test_list_instances_triggers_opencode_prune_for_visible_rows(self) -> None:
        app = TuiApp()
        rows = [
            EndpointInfo(
                name="vllm-qwen3",
                app_id="ap-123",
                state="running",
                backend=BackendType.VLLM,
                instance_name="qwen3",
            )
        ]
        with (
            patch("llm_launchpad.tui.app.ModalBackend.list_apps", return_value=rows),
            patch.object(app, "_sync_opencode") as sync_mock,
        ):
            app.list_instances()

        sync_mock.assert_called_once()
        synced_rows = sync_mock.call_args.kwargs["current_rows"]
        self.assertEqual(len(synced_rows), 1)
        self.assertEqual(synced_rows[0].name, "vllm-qwen3")

    def test_run_stop_syncs_opencode_with_removed_app_name(self) -> None:
        app = TuiApp()
        monitor = object()
        app._orchestrator = type(
            "FakeOrchestrator",
            (),
            {
                "stop_app": staticmethod(
                    lambda *_args, **_kwargs: [
                        OperationCompleteEvent(success=True, operation=OperationType.STOP),
                    ]
                ),
            },
        )()
        with (
            patch("llm_launchpad.tui.app._dispatch_event", return_value=None),
            patch.object(app, "_load_visible_launchpad_rows", return_value=[]),
            patch.object(app, "_sync_opencode") as sync_mock,
        ):
            app._run_stop(BackendType.VLLM, "vllm-qwen3", "vllm-qwen3", monitor=monitor)

        sync_mock.assert_called_once()
        self.assertEqual(sync_mock.call_args.kwargs["current_rows"], [])
        self.assertEqual(sync_mock.call_args.kwargs["remove_app_names"], ["vllm-qwen3"])
        self.assertIs(sync_mock.call_args.kwargs["monitor"], monitor)
        self.assertTrue(sync_mock.call_args.kwargs["emit_skipped"])

    def test_snapshot_persist_round_trip(self) -> None:
        with tempfile.TemporaryDirectory() as tmp:
            cache_path = Path(tmp) / "storage_snapshot.json"
            app = TuiApp()
            app._storage_cache_path = cache_path
            app._storage_snapshot_cache = None
            app._storage_snapshot_cached_at_epoch = 0.0

            snapshot = StorageSnapshot(
                llamacpp_models=[
                    StoredModelInfo(
                        backend=BackendType.LLAMACPP,
                        model_id="Qwen/Qwen2.5-Coder-7B-Instruct-GGUF",
                        revision="main",
                        quant="Q4_K_M",
                        size_bytes=1024,
                        file_count=1,
                        source_volume="huggingface-cache",
                        paths=["models/Qwen__Qwen2.5-Coder-7B-Instruct-GGUF/main/model.Q4_K_M.gguf"],
                        incomplete=True,
                    )
                ],
                vllm_models=[
                    StoredModelInfo(
                        backend=BackendType.VLLM,
                        model_id="Qwen/Qwen3-4B-Thinking-2507-FP8",
                        revision=None,
                        quant=None,
                        size_bytes=2048,
                        file_count=2,
                        source_volume="huggingface-cache",
                        paths=["hub/models--Qwen--Qwen3-4B-Thinking-2507-FP8"],
                    )
                ],
            )

            app._cache_storage_snapshot(snapshot)
            self.assertTrue(cache_path.exists())

            reloaded = TuiApp()
            reloaded._storage_cache_path = cache_path
            reloaded._storage_snapshot_cache = None
            reloaded._storage_snapshot_cached_at_epoch = 0.0
            reloaded._load_persisted_storage_cache()

            self.assertIsNotNone(reloaded._storage_snapshot_cache)
            loaded = reloaded._storage_snapshot_cache
            assert loaded is not None
            self.assertEqual(len(loaded.llamacpp_models), 1)
            self.assertEqual(len(loaded.vllm_models), 1)
            self.assertEqual(loaded.llamacpp_models[0].model_id, snapshot.llamacpp_models[0].model_id)
            self.assertEqual(loaded.vllm_models[0].model_id, snapshot.vllm_models[0].model_id)
            self.assertTrue(loaded.llamacpp_models[0].incomplete)

    def test_invalidate_storage_cache_removes_file(self) -> None:
        with tempfile.TemporaryDirectory() as tmp:
            cache_path = Path(tmp) / "storage_snapshot.json"
            app = TuiApp()
            app._storage_cache_path = cache_path
            snapshot = StorageSnapshot(llamacpp_models=[], vllm_models=[])
            app._cache_storage_snapshot(snapshot)
            self.assertTrue(cache_path.exists())

            app._invalidate_storage_cache()
            self.assertFalse(cache_path.exists())
            self.assertIsNone(app._storage_snapshot_cache)

    def test_load_persisted_storage_cache_handles_string_paths(self) -> None:
        with tempfile.TemporaryDirectory() as tmp:
            cache_path = Path(tmp) / "storage_snapshot.json"
            cache_path.write_text(
                json.dumps(
                    {
                        "cached_at_epoch": 123.0,
                        "snapshot": {
                            "llamacpp_models": [
                                {
                                    "backend": "llamacpp",
                                    "model_id": "legacy:model-Q4_K_M",
                                    "revision": None,
                                    "quant": "Q4_K_M",
                                    "size_bytes": 1024,
                                    "file_count": 1,
                                    "source_volume": "huggingface-cache",
                                    "paths": "/legacy/model-Q4_K_M.gguf",
                                    "incomplete": False,
                                }
                            ],
                            "vllm_models": [],
                        },
                    }
                )
            )

            app = TuiApp()
            app._storage_cache_path = cache_path
            app._storage_snapshot_cache = None
            app._storage_snapshot_cached_at_epoch = 0.0
            app._load_persisted_storage_cache()

            self.assertIsNotNone(app._storage_snapshot_cache)
            snapshot = app._storage_snapshot_cache
            assert snapshot is not None
            self.assertEqual(
                snapshot.llamacpp_models[0].paths,
                ["/legacy/model-Q4_K_M.gguf"],
            )


if __name__ == "__main__":
    unittest.main()



"""Core layer: business logic and orchestration.

This package owns configuration, Modal subprocess execution,
and high-level workflow coordination. It emits Protocol events
only and has zero UI dependencies.
"""

from .config import ConfigStore
from .backend import ModalBackend
from .modal_gpu import fetch_modal_gpu_types
from .orchestrator import Orchestrator

__all__ = ["ConfigStore", "ModalBackend", "Orchestrator", "fetch_modal_gpu_types"]



from __future__ import annotations

import unittest
from types import SimpleNamespace

from textual.app import App
from textual.screen import Screen
from textual.widgets import OptionList

from llm_launchpad.protocol.enums import BackendType
from llm_launchpad.protocol.models import EndpointInfo
from llm_launchpad.tui.screens.manage import (
    LogsParamsScreen,
    ManageScreen,
    StatusParamsScreen,
    StopParamsScreen,
)


class _TestApp(App[None]):
    def __init__(self) -> None:
        super().__init__()
        self.list_called = 0
        self.pushed: list[object] = []
        self.status_calls: list[tuple[BackendType, str | None, int, str | None, str | None]] = []
        self.logs_calls: list[tuple[BackendType, bool, str | None, str | None]] = []
        self.stop_calls: list[tuple[BackendType, str | None, str | None]] = []
        self.instances_by_backend: dict[BackendType, list[EndpointInfo]] = {
            BackendType.LLAMACPP: [],
            BackendType.VLLM: [],
        }

    def begin_list(self) -> None:
        self.list_called += 1

    def list_instances(self, _backend=None):  # type: ignore[no-untyped-def]
        if _backend is None:
            return self.instances_by_backend[BackendType.LLAMACPP] + self.instances_by_backend[BackendType.VLLM]
        return self.instances_by_backend.get(_backend, [])

    def push_screen(self, screen, *args, **kwargs):  # type: ignore[no-untyped-def]
        self.pushed.append(screen)
        return super().push_screen(screen, *args, **kwargs)

    def begin_status(  # type: ignore[no-untyped-def]
        self,
        backend,
        server_url=None,
        timeout=60,
        app_name=None,
        served_model_name=None,
    ) -> None:
        self.status_calls.append((backend, server_url, timeout, app_name, served_model_name))

    def begin_logs(  # type: ignore[no-untyped-def]
        self,
        backend,
        follow=True,
        app_name=None,
        app_id=None,
    ) -> None:
        self.logs_calls.append((backend, follow, app_name, app_id))

    def begin_stop(  # type: ignore[no-untyped-def]
        self,
        backend,
        app_name=None,
        app_id=None,
    ) -> None:
        self.stop_calls.append((backend, app_name, app_id))


class ManageScreenRoutingTests(unittest.IsolatedAsyncioTestCase):
    async def test_manage_screen_focuses_action_menu(self) -> None:
        app = _TestApp()
        async with app.run_test() as pilot:
            app.push_screen(ManageScreen())
            await pilot.pause()
            screen = app.screen
            assert isinstance(screen, ManageScreen)
            action_list = screen.query_one("#manage-action-list", OptionList)
            self.assertTrue(action_list.has_focus)
            highlighted = action_list.highlighted_option
            self.assertIsNotNone(highlighted)
            assert highlighted is not None
            self.assertEqual(highlighted.id, "list")

    async def test_manage_screen_action_routing(self) -> None:
        app = _TestApp()
        async with app.run_test() as pilot:
            app.push_screen(ManageScreen())
            await pilot.pause()
            screen = app.screen
            assert isinstance(screen, ManageScreen)

            for option_id in ["list", "status", "logs", "stop"]:
                screen.on_option_list_option_selected(
                    SimpleNamespace(option=SimpleNamespace(id=option_id))
                )
                await pilot.pause()

        self.assertEqual(app.list_called, 1)
        self.assertTrue(any(isinstance(s, StatusParamsScreen) for s in app.pushed))
        self.assertTrue(any(isinstance(s, LogsParamsScreen) for s in app.pushed))
        self.assertTrue(any(isinstance(s, StopParamsScreen) for s in app.pushed))

    async def test_status_params_menu_is_arrow_navigable(self) -> None:
        app = _TestApp()
        app.instances_by_backend[BackendType.LLAMACPP] = [
            EndpointInfo(name="llamacpp-alpha", app_id="ap-llama", state="running", backend=BackendType.LLAMACPP),
        ]
        app.instances_by_backend[BackendType.VLLM] = [
            EndpointInfo(name="vllm-beta", app_id="ap-vllm", state="deployed", backend=BackendType.VLLM),
        ]
        async with app.run_test() as pilot:
            app.push_screen(StatusParamsScreen())
            await pilot.pause()
            screen = app.screen
            assert isinstance(screen, StatusParamsScreen)
            instance_list = screen.query_one("#status-instance-list", OptionList)
            self.assertTrue(instance_list.has_focus)
            first = instance_list.highlighted_option
            self.assertIsNotNone(first)
            await pilot.press("down")
            await pilot.pause()
            second = instance_list.highlighted_option
            self.assertIsNotNone(second)
            assert first is not None
            assert second is not None
            self.assertNotEqual(first.id, second.id)

    async def test_logs_params_menu_is_arrow_navigable(self) -> None:
        app = _TestApp()
        app.instances_by_backend[BackendType.LLAMACPP] = [
            EndpointInfo(name="llamacpp-alpha", app_id="ap-llama", state="running", backend=BackendType.LLAMACPP),
        ]
        app.instances_by_backend[BackendType.VLLM] = [
            EndpointInfo(name="vllm-beta", app_id="ap-vllm", state="deployed", backend=BackendType.VLLM),
        ]
        async with app.run_test() as pilot:
            app.push_screen(LogsParamsScreen())
            await pilot.pause()
            screen = app.screen
            assert isinstance(screen, LogsParamsScreen)
            instance_list = screen.query_one("#logs-instance-list", OptionList)
            self.assertTrue(instance_list.has_focus)
            first = instance_list.highlighted_option
            self.assertIsNotNone(first)
            await pilot.press("down")
            await pilot.pause()
            second = instance_list.highlighted_option
            self.assertIsNotNone(second)
            assert first is not None
            assert second is not None
            self.assertNotEqual(first.id, second.id)

    async def test_stop_params_refreshes_instances_on_resume(self) -> None:
        app = _TestApp()
        app.instances_by_backend[BackendType.LLAMACPP] = [
            EndpointInfo(name="llamacpp-alpha", app_id="ap-llama", state="running", backend=BackendType.LLAMACPP),
        ]
        app.instances_by_backend[BackendType.VLLM] = [
            EndpointInfo(name="vllm-beta", app_id="ap-vllm", state="deployed", backend=BackendType.VLLM),
        ]

        async with app.run_test() as pilot:
            app.push_screen(StopParamsScreen())
            await pilot.pause()
            screen = app.screen
            assert isinstance(screen, StopParamsScreen)
            instance_list = screen.query_one("#stop-instance-list", OptionList)
            self.assertEqual(instance_list.option_count, 2)

            app.push_screen(Screen())
            await pilot.pause()

            app.instances_by_backend[BackendType.VLLM] = []
            app.pop_screen()
            await pilot.pause()

            screen = app.screen
            assert isinstance(screen, StopParamsScreen)
            instance_list = screen.query_one("#stop-instance-list", OptionList)
            self.assertEqual(instance_list.option_count, 1)
            highlighted = instance_list.highlighted_option
            self.assertIsNotNone(highlighted)
            assert highlighted is not None
            self.assertEqual(str(highlighted.id), "app-id:ap-llama")

    async def test_status_params_uses_selected_instance_web_url_and_served_model_name(self) -> None:
        app = _TestApp()
        app.instances_by_backend[BackendType.LLAMACPP] = [
            EndpointInfo(
                name="llamacpp-nanbeige",
                app_id="ap-llama",
                state="running",
                backend=BackendType.LLAMACPP,
                web_url="https://alice--llamacpp-nanbeige-serve-alpha-bravo.modal.run",
                served_model_name="Nanbeige4.1-3B-Q4_K_M-GGUF",
            ),
        ]

        async with app.run_test() as pilot:
            app.push_screen(StatusParamsScreen())
            await pilot.pause()
            screen = app.screen
            assert isinstance(screen, StatusParamsScreen)
            instance_list = screen.query_one("#status-instance-list", OptionList)
            highlighted = instance_list.highlighted_option
            self.assertIsNotNone(highlighted)
            assert highlighted is not None
            screen.on_option_list_option_selected(
                SimpleNamespace(option=highlighted, option_list=instance_list)
            )
            await pilot.pause()

        self.assertEqual(
            app.status_calls,
            [
                (
                    BackendType.LLAMACPP,
                    "https://alice--llamacpp-nanbeige-serve-alpha-bravo.modal.run",
                    60,
                    "llamacpp-nanbeige",
                    "Nanbeige4.1-3B-Q4_K_M-GGUF",
                )
            ],
        )

    async def test_logs_and_stop_use_exact_duplicate_app_id(self) -> None:
        duplicate_rows = [
            EndpointInfo(
                name="llamacpp-glm5-rtxpro",
                app_id="ap-first",
                state="ephemeral",
                backend=BackendType.LLAMACPP,
                instance_name="glm5-rtxpro",
            ),
            EndpointInfo(
                name="llamacpp-glm5-rtxpro",
                app_id="ap-second",
                state="ephemeral",
                backend=BackendType.LLAMACPP,
                instance_name="glm5-rtxpro",
            ),
        ]
        app = _TestApp()
        app.instances_by_backend[BackendType.LLAMACPP] = duplicate_rows

        async with app.run_test() as pilot:
            app.push_screen(LogsParamsScreen())
            await pilot.pause()
            screen = app.screen
            assert isinstance(screen, LogsParamsScreen)
            instance_list = screen.query_one("#logs-instance-list", OptionList)
            option = next(
                candidate for candidate in instance_list._options if str(candidate.id) == "app-id:ap-second"
            )
            screen.on_option_list_option_selected(
                SimpleNamespace(option=option, option_list=instance_list)
            )
            await pilot.pause()

            app.pop_screen()
            await pilot.pause()
            app.push_screen(StopParamsScreen())
            await pilot.pause()
            screen = app.screen
            assert isinstance(screen, StopParamsScreen)
            instance_list = screen.query_one("#stop-instance-list", OptionList)
            option = next(
                candidate for candidate in instance_list._options if str(candidate.id) == "app-id:ap-second"
            )
            screen.on_option_list_option_selected(
                SimpleNamespace(option=option, option_list=instance_list)
            )
            await pilot.pause()

        self.assertEqual(
            app.logs_calls,
            [(BackendType.LLAMACPP, True, "llamacpp-glm5-rtxpro", "ap-second")],
        )
        self.assertEqual(
            app.stop_calls,
            [(BackendType.LLAMACPP, "llamacpp-glm5-rtxpro", "ap-second")],
        )


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import unittest
from unittest.mock import patch

from textual.app import App
from textual.widgets import Log, Static

from llm_launchpad.protocol.enums import BackendType, DeploymentState, OperationType
from llm_launchpad.tui.screens.monitor import MonitorScreen
from llm_launchpad.tui.widgets.log_viewer import SelectableLog
from llm_launchpad.tui.workers import LogMessage, OperationDone, OperationError, StateChanged


class _TestApp(App[None]):
    pass


class MonitorScreenTests(unittest.IsolatedAsyncioTestCase):
    async def test_log_lines_are_appended_as_plain_text(self) -> None:
        app = _TestApp()
        async with app.run_test() as pilot:
            app.push_screen(MonitorScreen(title="Logs"))
            await pilot.pause()

            screen = app.screen
            assert isinstance(screen, MonitorScreen)
            self.assertIsInstance(screen.log_viewer.log_widget, Log)
            screen.on_log_message(LogMessage("first line"))
            screen.on_log_message(LogMessage("stderr line", stream="stderr"))
            await pilot.pause()

            content = "\n".join(screen.log_viewer.log_widget.lines)
            self.assertIn("first line", content)
            self.assertIn("stderr | stderr line", content)
            title = screen.query_one("#monitor-title", Static)
            self.assertIn(
                "ctrl+shift+c copy  y fallback  ctrl+t terminal copy  ctrl+c exits  ctrl+l clear",
                str(title.content),
            )

    async def test_title_mentions_terminal_selection_when_mouse_disabled(self) -> None:
        app = _TestApp()
        app.mouse_enabled = False
        async with app.run_test() as pilot:
            app.push_screen(MonitorScreen(title="Logs"))
            await pilot.pause()

            screen = app.screen
            assert isinstance(screen, MonitorScreen)
            title = screen.query_one("#monitor-title", Static)
            self.assertIn(
                "terminal selection mode  use your terminal copy shortcut  ctrl+t mouse  ctrl+c exits",
                str(title.content),
            )

    async def test_log_lines_strip_ansi_escape_sequences(self) -> None:
        app = _TestApp()
        async with app.run_test() as pilot:
            app.push_screen(MonitorScreen(title="Logs"))
            await pilot.pause()

            screen = app.screen
            assert isinstance(screen, MonitorScreen)
            screen.on_log_message(LogMessage("\x1b[0;36m(APIServer pid=4)\x1b[0;0m INFO booted"))
            await pilot.pause()

            content = "\n".join(screen.log_viewer.log_widget.lines)
            self.assertIn("(APIServer pid=4) INFO booted", content)
            self.assertNotIn("\x1b[0;36m", content)
            self.assertNotIn("\x1b[0;0m", content)

    async def test_clear_action_empties_log(self) -> None:
        app = _TestApp()
        async with app.run_test() as pilot:
            app.push_screen(MonitorScreen(title="Logs"))
            await pilot.pause()

            screen = app.screen
            assert isinstance(screen, MonitorScreen)
            screen.on_log_message(LogMessage("line to clear"))
            await pilot.pause()
            self.assertGreater(screen.log_viewer.log_widget.line_count, 0)

            screen.action_clear_log()
            await pilot.pause()
            self.assertEqual(screen.log_viewer.log_widget.line_count, 0)

    async def test_operation_messages_render_without_markup(self) -> None:
        app = _TestApp()
        async with app.run_test() as pilot:
            app.push_screen(MonitorScreen(title="Deploy"))
            await pilot.pause()

            screen = app.screen
            assert isinstance(screen, MonitorScreen)
            screen.on_operation_done(
                OperationDone(
                    operation=OperationType.DEPLOY,
                    success=False,
                    exit_code=9,
                    detail="backend error",
                )
            )
            screen.on_operation_error(OperationError("worker failed"))
            await pilot.pause()

            content = "\n".join(screen.log_viewer.log_widget.lines)
            self.assertIn("Operation failed (exit code 9).", content)
            self.assertIn("Detail: backend error", content)
            self.assertIn("Press esc or q to return.", content)
            self.assertIn("Error: worker failed", content)
            self.assertNotIn("[red", content)
            self.assertNotIn("[green", content)
            self.assertNotIn("[dim]", content)

    async def test_summary_mode_normalizes_backend_log_lines(self) -> None:
        app = _TestApp()
        async with app.run_test() as pilot:
            app.push_screen(
                MonitorScreen(
                    title="Deploy",
                    deploy_backend=BackendType.VLLM,
                    summarize_backend_logs=True,
                    show_debug_logs=False,
                )
            )
            await pilot.pause()

            screen = app.screen
            assert isinstance(screen, MonitorScreen)
            screen.on_state_changed(
                StateChanged(
                    state=DeploymentState.WARMING_UP,
                    operation=OperationType.WARMUP,
                    detail="warming",
                )
            )
            screen.on_log_message(LogMessage("(APIServer pid=4) INFO 02-25 03:08:40 vLLM API server version 0.13.0"))
            await pilot.pause()

            content = "\n".join(screen.log_viewer.log_widget.lines)
            self.assertIn("Log view: summary (normalized milestones; raw backend logs hidden)", content)
            self.assertIn("Starting server", content)
            self.assertNotIn("vLLM API server version", content)

    async def test_debug_mode_preserves_raw_backend_log_lines(self) -> None:
        app = _TestApp()
        async with app.run_test() as pilot:
            app.push_screen(
                MonitorScreen(
                    title="Deploy",
                    deploy_backend=BackendType.VLLM,
                    summarize_backend_logs=True,
                    show_debug_logs=True,
                )
            )
            await pilot.pause()

            screen = app.screen
            assert isinstance(screen, MonitorScreen)
            raw_line = "(APIServer pid=4) INFO 02-25 03:08:40 vLLM API server version 0.13.0"
            screen.on_log_message(LogMessage(raw_line))
            await pilot.pause()

            content = "\n".join(screen.log_viewer.log_widget.lines)
            self.assertIn(raw_line, content)
            self.assertNotIn("Log view: summary (normalized milestones; raw backend logs hidden)", content)

    async def test_debug_mode_preserves_ansi_escape_sequences(self) -> None:
        app = _TestApp()
        async with app.run_test() as pilot:
            app.push_screen(
                MonitorScreen(
                    title="Deploy",
                    deploy_backend=BackendType.VLLM,
                    summarize_backend_logs=True,
                    show_debug_logs=True,
                )
            )
            await pilot.pause()

            screen = app.screen
            assert isinstance(screen, MonitorScreen)
            raw_line = "\x1b[0;36m(APIServer pid=4)\x1b[0;0m INFO booted"
            with patch.object(screen.log_viewer, "write_line") as write_line:
                screen.on_log_message(LogMessage(raw_line))

            write_line.assert_called_once_with(raw_line)

    async def test_summary_mode_failure_appends_debug_hint(self) -> None:
        app = _TestApp()
        async with app.run_test() as pilot:
            app.push_screen(
                MonitorScreen(
                    title="Deploy",
                    deploy_backend=BackendType.LLAMACPP,
                    summarize_backend_logs=True,
                    show_debug_logs=False,
                )
            )
            await pilot.pause()

            screen = app.screen
            assert isinstance(screen, MonitorScreen)
            screen.on_operation_done(
                OperationDone(
                    operation=OperationType.WARMUP,
                    success=False,
                    exit_code=1,
                    detail="timed out",
                )
            )
            await pilot.pause()

            content = "\n".join(screen.log_viewer.log_widget.lines)
            self.assertIn('Tip: re-run with "Show debug logs" enabled to see full backend logs.', content)

    async def test_log_widget_is_selectable_log(self) -> None:
        """LogViewer should use SelectableLog (not plain Log) for line selection."""
        app = _TestApp()
        async with app.run_test() as pilot:
            app.push_screen(MonitorScreen(title="Logs"))
            await pilot.pause()

            screen = app.screen
            assert isinstance(screen, MonitorScreen)
            self.assertIsInstance(screen.log_viewer.log_widget, SelectableLog)

    async def test_double_click_line_selection_copies_line_to_clipboard(self) -> None:
        app = _TestApp()
        async with app.run_test() as pilot:
            app.push_screen(MonitorScreen(title="Logs"))
            await pilot.pause()

            screen = app.screen
            assert isinstance(screen, MonitorScreen)
            screen.on_log_message(LogMessage("first line"))
            await pilot.pause()

            copied = screen.log_viewer.log_widget._select_line(0)
            self.assertTrue(copied)
            await pilot.pause()
            await pilot.pause()

            self.assertEqual(app.clipboard, "first line")

    def test_copy_binding_includes_terminal_safe_variants(self) -> None:
        copy_binding = next(b for b in MonitorScreen.BINDINGS if b.action == "copy_text")
        self.assertEqual(copy_binding.key, "y")

    def test_direct_clipboard_binding_keeps_terminal_copy_aliases(self) -> None:
        copy_binding = next(
            b for b in MonitorScreen.BINDINGS if b.action == "copy_text_to_clipboard"
        )
        self.assertEqual(
            copy_binding.key,
            "ctrl+shift+c,super+c,meta+c,cmd+c,command+c",
        )


if __name__ == "__main__":
    unittest.main()



"""Settings persistence and configuration management."""

from __future__ import annotations

import json
from pathlib import Path
from typing import Any, Dict

from ..protocol.models import LaunchpadSettings

SETTINGS_DIR = Path.home() / ".llm_launchpad"
SETTINGS_PATH = SETTINGS_DIR / "settings.json"


class ConfigStore:
    """Load and save user settings from ``~/.llm_launchpad/settings.json``."""

    def __init__(self, path: Path | None = None) -> None:
        self.path = path or SETTINGS_PATH

    def load(self) -> LaunchpadSettings:
        """Load settings, returning defaults when missing or corrupt."""
        if self.path.exists():
            try:
                raw: Dict[str, Any] = json.loads(self.path.read_text())
                return LaunchpadSettings.from_dict(raw)
            except Exception:
                return LaunchpadSettings()
        return LaunchpadSettings()

    def save(self, settings: LaunchpadSettings) -> None:
        """Persist settings. Failures are silently ignored."""
        try:
            self.path.parent.mkdir(parents=True, exist_ok=True)
            self.path.write_text(json.dumps(settings.to_dict(), indent=2))
        except Exception:
            pass



"""Log viewer widget: selectable plain-text Log wrapper."""

from __future__ import annotations

from textual import events
from textual.app import ComposeResult
from textual.containers import Vertical
from textual.geometry import Offset
from textual.selection import Selection
from textual.widgets import Log


class SelectableLog(Log):
    """Log with line-level double-click selection.

    * **Double-click** selects the entire clicked line.
    * **Triple-click** selects all log content.
    * Single click / drag selection is unchanged.
    """

    def _select_line(self, line_index: int) -> bool:
        """Select a single line."""
        if not (0 <= line_index < self.line_count):
            return False
        line_text = self.lines[line_index]
        start = Offset(0, line_index)
        end = Offset(len(line_text), line_index)
        self.screen.selections = {
            self: Selection(start, end)
        }
        return True

    def _select_all(self) -> None:
        """Select all log content."""
        self.text_select_all()

    async def _on_click(self, event: events.Click) -> None:
        if event.widget is self:
            if (
                self.allow_select
                and self.screen.allow_select
                and self.app.ALLOW_SELECT
            ):
                if event.chain == 2:
                    # Select only the clicked line (not all text).
                    # prevent_default() stops Textual from also calling
                    # the parent Log._on_click which would text_select_all().
                    line_index = int(self.scroll_y) + event.y
                    if self._select_line(line_index):
                        event.prevent_default()
                        return
                elif event.chain == 3:
                    self._select_all()
                    event.prevent_default()
                    return
        await self.broker_event("click", event)


class LogViewer(Vertical):
    """Scrollable log output panel with auto-scroll.

    Wraps a selectable ``SelectableLog`` and exposes a simple ``write_line``
    API.
    """

    DEFAULT_CSS = """
    LogViewer {
        height: 1fr;
        border: solid #17321e;
        background: #0a0f0b;
        padding: 0 1;
    }
    LogViewer Log {
        height: 1fr;
    }
    """

    def compose(self) -> ComposeResult:
        yield SelectableLog(highlight=True, auto_scroll=True, id="log-output")

    @property
    def log_widget(self) -> SelectableLog:
        return self.query_one("#log-output", SelectableLog)

    def write_line(self, line: str, style: str = "") -> None:
        """Append a line and auto-scroll to bottom.

        ``style`` is accepted for compatibility with existing call sites.
        """
        _ = style
        self.log_widget.write_line(line)

    def clear(self) -> None:
        self.log_widget.clear()



from __future__ import annotations

import unittest
from types import SimpleNamespace
from unittest.mock import patch

from textual.app import App, ComposeResult
from textual.selection import SELECT_ALL
from textual.widgets import DataTable, Input, OptionList, Static
from textual.widgets.option_list import Option

from llm_launchpad.tui.app import TuiApp, _osc_52_sequence, _screen_passthrough_sequence, _tmux_passthrough_sequence
from llm_launchpad.tui.screens.copy_enabled import CopyEnabledScreen
from llm_launchpad.tui.screens.deploy import (
    BackendSelectScreen,
    LlamaCppDeployScreen,
    VllmDeployScreen,
)
from llm_launchpad.tui.screens.main_menu import MainMenuScreen
from llm_launchpad.tui.screens.manage import (
    LogsParamsScreen,
    ManageScreen,
    StatusParamsScreen,
    StopParamsScreen,
)
from llm_launchpad.tui.screens.monitor import MonitorScreen
from llm_launchpad.tui.screens.settings import SettingsScreen
from llm_launchpad.tui.screens.storage import StorageScreen


class _StaticSelectionScreen(CopyEnabledScreen):
    def compose(self) -> ComposeResult:
        yield Static("[bold]Hello[/bold] world", id="copy-static")


class _OptionListFallbackScreen(CopyEnabledScreen):
    def compose(self) -> ComposeResult:
        yield OptionList(
            Option("  First option", id="first"),
            Option("  Second option", id="second"),
            id="copy-options",
        )

    def on_mount(self) -> None:
        options = self.query_one("#copy-options", OptionList)
        options.highlighted = 1
        options.focus()


class _DataTableFallbackScreen(CopyEnabledScreen):
    def compose(self) -> ComposeResult:
        yield DataTable(id="copy-table")

    def on_mount(self) -> None:
        table = self.query_one("#copy-table", DataTable)
        table.cursor_type = "row"
        table.add_columns("backend", "model", "revision")
        table.add_row("llamacpp", "Qwen/Qwen2.5-Coder-7B-Instruct-GGUF", "main")
        table.add_row("vllm", "Qwen/Qwen3-4B-Thinking-2507-FP8", "-")
        table.move_cursor(row=1, column=0, animate=False)
        table.focus()


class _InputFallbackScreen(CopyEnabledScreen):
    def compose(self) -> ComposeResult:
        yield Input(value="Qwen/Qwen3-8B", id="copy-input")

    def on_mount(self) -> None:
        self.query_one("#copy-input", Input).focus()


class _NoCopyFallbackScreen(CopyEnabledScreen):
    def compose(self) -> ComposeResult:
        yield Static("Nothing focusable/copyable here")


class _CopyTestApp(App[None]):
    pass


class _CopyKeyApp(App[None]):
    def __init__(self) -> None:
        super().__init__()
        self.help_quit_called = False

    def action_help_quit(self) -> None:
        self.help_quit_called = True


class _QuitCaptureApp(TuiApp):
    def __init__(self) -> None:
        super().__init__()
        self.notifications: list[tuple[object, object, object, object]] = []
        self.quit_calls = 0

    def notify(
        self,
        message: object,
        *,
        title: object | None = None,
        severity: str = "information",
        timeout: float | None = None,
        **kwargs: object,
    ) -> None:
        self.notifications.append((message, title, severity, timeout))

    async def action_quit(self) -> None:
        self.quit_calls += 1


class _MouseDriverStub:
    def __init__(self, mouse: bool) -> None:
        self._mouse = mouse
        self.enable_calls = 0
        self.disable_calls = 0

    def _enable_mouse_support(self) -> None:
        self.enable_calls += 1

    def _disable_mouse_support(self) -> None:
        self.disable_calls += 1


class CopyEnabledScreenTests(unittest.IsolatedAsyncioTestCase):
    async def test_selection_copy_from_static_writes_to_clipboard(self) -> None:
        app = _CopyTestApp()
        async with app.run_test() as pilot:
            app.push_screen(_StaticSelectionScreen())
            await pilot.pause()

            screen = app.screen
            assert isinstance(screen, _StaticSelectionScreen)
            target = screen.query_one("#copy-static", Static)
            screen.selections = {target: SELECT_ALL}
            await pilot.pause()

            screen.action_copy_text()
            self.assertEqual(app.clipboard, "Hello world")

    async def test_option_list_fallback_copies_highlighted_option(self) -> None:
        app = _CopyTestApp()
        async with app.run_test() as pilot:
            app.push_screen(_OptionListFallbackScreen())
            await pilot.pause()

            screen = app.screen
            assert isinstance(screen, _OptionListFallbackScreen)
            screen.action_copy_text()
            self.assertIn("Second option", app.clipboard)

    async def test_data_table_fallback_copies_highlighted_row_as_tsv(self) -> None:
        app = _CopyTestApp()
        async with app.run_test() as pilot:
            app.push_screen(_DataTableFallbackScreen())
            await pilot.pause()

            screen = app.screen
            assert isinstance(screen, _DataTableFallbackScreen)
            screen.action_copy_text()
            self.assertEqual(
                app.clipboard,
                "vllm\tQwen/Qwen3-4B-Thinking-2507-FP8\t-",
            )

    async def test_input_fallback_copies_full_value(self) -> None:
        app = _CopyTestApp()
        async with app.run_test() as pilot:
            app.push_screen(_InputFallbackScreen())
            await pilot.pause()

            screen = app.screen
            assert isinstance(screen, _InputFallbackScreen)
            screen.action_copy_text()
            self.assertEqual(app.clipboard, "Qwen/Qwen3-8B")

    async def test_input_ctrl_c_uses_input_native_copy_behavior(self) -> None:
        app = _CopyKeyApp()
        async with app.run_test() as pilot:
            app.push_screen(_InputFallbackScreen())
            await pilot.pause()

            screen = app.screen
            assert isinstance(screen, _InputFallbackScreen)
            input_widget = screen.query_one("#copy-input", Input)
            input_widget.action_select_all()
            await pilot.pause()

            await pilot.press("ctrl+c")
            await pilot.pause()

            self.assertFalse(app.help_quit_called)
            self.assertEqual(app.clipboard, "Qwen/Qwen3-8B")

    async def test_y_alias_triggers_copy_action(self) -> None:
        app = _CopyTestApp()
        async with app.run_test() as pilot:
            app.push_screen(_StaticSelectionScreen())
            await pilot.pause()

            screen = app.screen
            assert isinstance(screen, _StaticSelectionScreen)
            target = screen.query_one("#copy-static", Static)
            screen.selections = {target: SELECT_ALL}
            await pilot.pause()

            await pilot.press("y")
            await pilot.pause()

            self.assertEqual(app.clipboard, "Hello world")

    async def test_selection_change_auto_copies_selection(self) -> None:
        app = _CopyTestApp()
        async with app.run_test() as pilot:
            app.push_screen(_StaticSelectionScreen())
            await pilot.pause()

            screen = app.screen
            assert isinstance(screen, _StaticSelectionScreen)
            target = screen.query_one("#copy-static", Static)
            target.text_select_all()
            await pilot.pause()
            await pilot.pause()

            self.assertEqual(app.clipboard, "Hello world")

    async def test_ctrl_shift_c_binding_copies_selected_text_to_clipboard(self) -> None:
        app = _CopyTestApp()
        async with app.run_test() as pilot:
            app.push_screen(_StaticSelectionScreen())
            await pilot.pause()

            screen = app.screen
            assert isinstance(screen, _StaticSelectionScreen)
            target = screen.query_one("#copy-static", Static)
            screen.selections = {target: SELECT_ALL}
            await pilot.pause()

            await pilot.press("ctrl+shift+c")
            await pilot.pause()

            self.assertEqual(app.clipboard, "Hello world")

    async def test_super_c_binding_copies_selected_text_to_clipboard(self) -> None:
        app = _CopyTestApp()
        async with app.run_test() as pilot:
            app.push_screen(_StaticSelectionScreen())
            await pilot.pause()

            screen = app.screen
            assert isinstance(screen, _StaticSelectionScreen)
            target = screen.query_one("#copy-static", Static)
            screen.selections = {target: SELECT_ALL}
            await pilot.pause()

            await pilot.press("super+c")
            await pilot.pause()

            self.assertEqual(app.clipboard, "Hello world")

    async def test_direct_clipboard_action_copies_to_clipboard(self) -> None:
        app = _CopyTestApp()
        async with app.run_test() as pilot:
            app.push_screen(_StaticSelectionScreen())
            await pilot.pause()

            screen = app.screen
            assert isinstance(screen, _StaticSelectionScreen)
            target = screen.query_one("#copy-static", Static)
            screen.selections = {target: SELECT_ALL}
            await pilot.pause()

            screen.action_copy_text_to_clipboard()

            self.assertEqual(app.clipboard, "Hello world")

    async def test_selected_text_hook_does_not_fall_back_without_selection(self) -> None:
        app = _CopyTestApp()
        async with app.run_test() as pilot:
            app.push_screen(_OptionListFallbackScreen())
            await pilot.pause()

            screen = app.screen
            assert isinstance(screen, _OptionListFallbackScreen)
            self.assertIsNone(screen._selected_text_for_copy())
            self.assertEqual(app.clipboard, "")


class TuiAppClipboardTests(unittest.TestCase):
    def test_copy_to_clipboard_uses_pbcopy_on_darwin(self) -> None:
        app = TuiApp()
        with (
            patch("llm_launchpad.tui.app.sys.platform", "darwin"),
            patch("llm_launchpad.tui.app.subprocess.run") as run_mock,
        ):
            app.copy_to_clipboard("copied text")
        self.assertEqual(app.clipboard, "copied text")
        run_mock.assert_called_once_with(
            ["pbcopy"],
            input=b"copied text",
            check=True,
            timeout=2,
        )

    def test_copy_to_clipboard_writes_tmux_passthrough_sequence(self) -> None:
        writes: list[str] = []
        app = TuiApp()
        app._driver = SimpleNamespace(write=writes.append)

        with patch.dict("llm_launchpad.tui.app.os.environ", {"TMUX": "/tmp/tmux-1000/default,123,0"}):
            app.copy_to_clipboard("copied text")

        self.assertEqual(app.clipboard, "copied text")
        self.assertEqual(
            writes,
            [
                _osc_52_sequence("copied text"),
                _tmux_passthrough_sequence("copied text"),
            ],
        )

    def test_copy_to_clipboard_writes_screen_passthrough_sequence(self) -> None:
        writes: list[str] = []
        app = TuiApp()
        app._driver = SimpleNamespace(write=writes.append)

        with patch.dict("llm_launchpad.tui.app.os.environ", {"TERM": "screen-256color"}, clear=True):
            app.copy_to_clipboard("copied text")

        self.assertEqual(app.clipboard, "copied text")
        self.assertEqual(
            writes,
            [
                _osc_52_sequence("copied text"),
                _screen_passthrough_sequence("copied text"),
            ],
        )


class TuiAppQuitTests(unittest.IsolatedAsyncioTestCase):
    def test_app_bindings_do_not_include_q_quit(self) -> None:
        self.assertFalse(any(binding.key == "q" for binding in TuiApp.BINDINGS))

    async def test_ctrl_c_first_press_warns_before_quitting(self) -> None:
        app = _QuitCaptureApp()

        with patch("llm_launchpad.tui.app.time.monotonic", return_value=10.0):
            await app.action_request_quit()

        self.assertEqual(app.quit_calls, 0)
        self.assertEqual(
            app.notifications,
            [("Hit CTRL+C again to exit", "Exit llm-launchpad?", "warning", 10.0)],
        )

    async def test_ctrl_c_second_press_within_window_quits(self) -> None:
        app = _QuitCaptureApp()

        with patch("llm_launchpad.tui.app.time.monotonic", side_effect=[10.0, 19.5]):
            await app.action_request_quit()
            await app.action_request_quit()

        self.assertEqual(app.quit_calls, 1)

    async def test_ctrl_c_after_window_warns_again(self) -> None:
        app = _QuitCaptureApp()

        with patch("llm_launchpad.tui.app.time.monotonic", side_effect=[10.0, 20.5]):
            await app.action_request_quit()
            await app.action_request_quit()

        self.assertEqual(app.quit_calls, 0)
        self.assertEqual(len(app.notifications), 2)


class TuiAppMouseModeTests(unittest.TestCase):
    def test_toggle_mouse_mode_enables_driver_mouse_support(self) -> None:
        app = TuiApp(mouse_enabled=False)
        driver = _MouseDriverStub(mouse=False)
        app._driver = driver

        app.action_toggle_mouse_mode()

        self.assertTrue(app.mouse_enabled)
        self.assertTrue(driver._mouse)
        self.assertEqual(driver.enable_calls, 1)
        self.assertEqual(driver.disable_calls, 0)

    def test_toggle_mouse_mode_disables_driver_mouse_support(self) -> None:
        app = TuiApp(mouse_enabled=True)
        driver = _MouseDriverStub(mouse=True)
        app._driver = driver

        app.action_toggle_mouse_mode()

        self.assertFalse(app.mouse_enabled)
        self.assertFalse(driver._mouse)
        self.assertEqual(driver.enable_calls, 0)
        self.assertEqual(driver.disable_calls, 1)


class CopyEnabledScreenInheritanceTests(unittest.TestCase):
    def test_all_tui_screens_subclass_copy_enabled_screen(self) -> None:
        screen_classes = [
            SettingsScreen,
            MainMenuScreen,
            MonitorScreen,
            BackendSelectScreen,
            LlamaCppDeployScreen,
            VllmDeployScreen,
            StorageScreen,
            ManageScreen,
            StatusParamsScreen,
            LogsParamsScreen,
            StopParamsScreen,
        ]
        for screen_class in screen_classes:
            self.assertTrue(issubclass(screen_class, CopyEnabledScreen), screen_class.__name__)


if __name__ == "__main__":
    unittest.main()



"""Fetch Modal GPU types from the Modal docs page."""

from __future__ import annotations

from dataclasses import dataclass
import html
import re
from typing import Final

MODAL_GPU_GUIDE_URL: Final[str] = "https://modal.com/docs/guide/gpu"
MODAL_PRICING_URL: Final[str] = "https://modal.com/pricing"
_REQUEST_HEADERS: Final[dict[str, str]] = {
    "User-Agent": (
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) "
        "AppleWebKit/537.36 (KHTML, like Gecko) "
        "Chrome/131.0.0.0 Safari/537.36"
    ),
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
}

_GPU_TOKEN_RE: Final[re.Pattern[str]] = re.compile(
    r"(?<![A-Za-z0-9_-])"
    r"([A-Z][A-Z0-9]*(?:-[A-Z0-9]+)*(?:[+!])?)"
    r"(?::\d+)?"
    r"(?![A-Za-z0-9_-])"
)
_VALID_GPU_RE: Final[re.Pattern[str]] = re.compile(
    r"(?=.*\d)[A-Z][A-Z0-9]*(?:-[A-Z0-9]+)*(?:[+!])?"
)
_HTML_TAG_RE: Final[re.Pattern[str]] = re.compile(r"<[^>]+>")
_PRICING_ROW_RE: Final[re.Pattern[str]] = re.compile(
    r"(NVIDIA\s+[A-Z0-9 ,+!-]+?)\s+\$([0-9]+(?:\.[0-9]+)?)\s*/\s*(SEC(?:OND)?|HR|HOUR)\b",
    re.IGNORECASE,
)
_WHITESPACE_RE: Final[re.Pattern[str]] = re.compile(r"\s+")
_SECONDS_PER_HOUR: Final[float] = 3600.0


@dataclass(frozen=True)
class ModalGpuSpec:
    """GPU value plus optional Modal pricing metadata."""

    value: str
    price_per_hour_usd: float | None = None


def fetch_modal_gpu_types(
    url: str = MODAL_GPU_GUIDE_URL,
    timeout: float = 10.0,
) -> list[str]:
    """Fetch currently documented Modal GPU type values."""
    page_text = _fetch_modal_page_text(url=url, timeout=timeout)
    gpu_types = _parse_modal_gpu_types(page_text)
    if not gpu_types:
        raise RuntimeError(
            "Could not parse GPU types from Modal docs page. The page format may have changed."
        )
    return gpu_types


def fetch_modal_gpu_catalog(
    gpu_guide_url: str = MODAL_GPU_GUIDE_URL,
    pricing_url: str = MODAL_PRICING_URL,
    timeout: float = 10.0,
) -> list[ModalGpuSpec]:
    """Fetch documented Modal GPU values with current base hourly pricing when available."""
    from .backend import ModalBackend

    gpu_types = fetch_modal_gpu_types(url=gpu_guide_url, timeout=timeout)

    pricing_by_gpu: dict[str, float] = {}
    try:
        pricing_page_text = _fetch_modal_page_text(url=pricing_url, timeout=timeout)
    except Exception:
        if ModalBackend.is_shutting_down():
            raise
    else:
        pricing_by_gpu = _parse_modal_gpu_pricing(pricing_page_text)

    return [
        ModalGpuSpec(
            value=gpu_type,
            price_per_hour_usd=pricing_by_gpu.get(gpu_type),
        )
        for gpu_type in gpu_types
    ]


def _parse_modal_gpu_types(page_text: str) -> list[str]:
    """Parse GPU identifiers from Modal docs HTML or markdown text."""
    content = _strip_html_tags(_slice_gpu_type_section(page_text))
    raw_tokens = _GPU_TOKEN_RE.findall(content)
    if not raw_tokens:
        raw_tokens = _GPU_TOKEN_RE.findall(_strip_html_tags(page_text))

    parsed: list[str] = []
    seen: set[str] = set()
    for raw in raw_tokens:
        token = _normalize_gpu_token(raw)
        if token is None or token in seen:
            continue
        seen.add(token)
        parsed.append(token)
    return parsed


def _parse_modal_gpu_pricing(page_text: str) -> dict[str, float]:
    """Parse base hourly Modal GPU pricing from the pricing page."""
    content = _normalize_whitespace(_strip_html_tags(_slice_gpu_pricing_section(page_text)))
    pricing: dict[str, float] = {}
    for raw_name, raw_price, raw_unit in _PRICING_ROW_RE.findall(content):
        token = _normalize_pricing_gpu_token(raw_name)
        if token is None:
            continue
        hourly_price = float(raw_price)
        if raw_unit.strip().upper().startswith("SEC"):
            hourly_price *= _SECONDS_PER_HOUR
        pricing[token] = hourly_price
    return _with_pricing_aliases(pricing)


def _slice_gpu_type_section(page_text: str) -> str:
    lower = page_text.lower()
    start = lower.find("specifying gpu type")
    if start == -1:
        start = lower.find("supports the following values for this parameter")
    if start == -1:
        return page_text

    end = lower.find("specifying gpu count", start)
    if end == -1:
        end = lower.find("gpu fallbacks", start)
    if end == -1:
        end = len(page_text)
    return page_text[start:end]


def _slice_gpu_pricing_section(page_text: str) -> str:
    lower = page_text.lower()
    start = lower.find("gpu tasks")
    if start == -1:
        return page_text

    end_candidates = [
        lower.find("cpu", start),
        lower.find("pricing plans", start),
        lower.find("modal sandbox", start),
    ]
    valid_end_candidates = [idx for idx in end_candidates if idx != -1]
    end = min(valid_end_candidates) if valid_end_candidates else len(page_text)
    return page_text[start:end]


def _normalize_gpu_token(value: str) -> str | None:
    token = value.strip().strip("`'\"").upper()
    if not token:
        return None
    if not _VALID_GPU_RE.fullmatch(token):
        return None
    return token


def _normalize_pricing_gpu_token(value: str) -> str | None:
    normalized = _normalize_whitespace(_strip_html_tags(value)).upper()
    if normalized.startswith("NVIDIA "):
        normalized = normalized[len("NVIDIA ") :].strip()

    a100_match = re.fullmatch(r"([A-Z0-9]+),\s*(\d+)\s*GB", normalized)
    if a100_match is not None:
        return f"{a100_match.group(1)}-{a100_match.group(2)}GB"

    collapsed = normalized.replace(",", "").replace(" ", "-")
    collapsed = re.sub(r"-{2,}", "-", collapsed)
    if not collapsed:
        return None
    if not _VALID_GPU_RE.fullmatch(collapsed):
        return None
    return collapsed


def _with_pricing_aliases(pricing: dict[str, float]) -> dict[str, float]:
    aliased = dict(pricing)
    alias_map = {
        "A100-40GB": ("A100",),
        "H100": ("H100!",),
        "B200": ("B200+",),
    }
    for source, aliases in alias_map.items():
        price = aliased.get(source)
        if price is None:
            continue
        for alias in aliases:
            aliased.setdefault(alias, price)
    return aliased


def _strip_html_tags(value: str) -> str:
    unescaped = html.unescape(value)
    return _HTML_TAG_RE.sub(" ", unescaped)


def _normalize_whitespace(value: str) -> str:
    return _WHITESPACE_RE.sub(" ", value).strip()


def _fetch_modal_page_text(url: str, timeout: float) -> str:
    from .backend import ModalBackend

    try:
        import requests
    except Exception as exc:
        raise RuntimeError(
            "requests is required to fetch Modal GPU metadata. Install with: pip install requests"
        ) from exc

    try:
        response = requests.get(url, timeout=timeout, headers=_REQUEST_HEADERS)
    except Exception as exc:
        if ModalBackend.is_shutting_down():
            raise RuntimeError("Shutdown requested") from exc
        raise RuntimeError(f"Failed to fetch Modal metadata page at {url}: {exc}") from exc

    if response.status_code >= 400:
        raise RuntimeError(f"Modal metadata page returned HTTP {response.status_code} for {url}")
    return response.text



"""Storage management screen for cached models across backends."""

from __future__ import annotations

from textual import events
from textual.actions import SkipAction
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Vertical, VerticalScroll
from textual.widget import Widget
from textual.widgets import DataTable, Footer, Input, OptionList, Static
from textual.widgets.option_list import Option

from ...protocol.enums import BackendType
from ...protocol.models import StorageSnapshot, StoredModelInfo
from ..navigation import (
    first_enabled_option_index,
    is_focusable_for_navigation,
    move_focus_across_option_lists,
    move_focus_across_widgets,
)
from ..workers import StorageFailed, StorageLoaded
from .copy_enabled import CopyEnabledScreen


def _human_bytes(size_bytes: int) -> str:
    size = float(size_bytes)
    units = ["B", "KB", "MB", "GB", "TB"]
    for unit in units:
        if size < 1024.0 or unit == units[-1]:
            return f"{size:.1f} {unit}"
        size /= 1024.0
    return f"{size_bytes} B"


def _model_label(row: StoredModelInfo) -> str:
    if row.incomplete:
        return f"{row.model_id} (INCOMPLETE)"
    return row.model_id


def _is_focusable_for_arrow_navigation(widget: Widget) -> bool:
    return is_focusable_for_navigation(widget, check_size=True)


class StorageScreen(CopyEnabledScreen):
    """View and pre-download backend model caches."""

    BINDINGS = [
        Binding("up", "navigate_option_list_up", show=False, priority=True),
        Binding("down", "navigate_option_list_down", show=False, priority=True),
        Binding("escape", "pop_screen", "Back", show=True),
        Binding("r", "refresh_storage", "Refresh", show=True),
        Binding("p", "predownload_selected", "Pre-download", show=True),
        Binding("x", "delete_selected_model", "Delete", show=True),
    ]
    NAVIGATION_ORDER = (
        "storage-backend-filter",
        "storage-action-list",
        "storage-table",
        "storage-model-id",
        "storage-model-backend",
        "storage-model-quant",
        "storage-model-revision",
    )

    def __init__(self, initial_backend: BackendType | None = None) -> None:
        super().__init__()
        self._initial_backend = initial_backend

    def compose(self) -> ComposeResult:
        with VerticalScroll(id="menu-container"):
            yield Static("[bold #7bf168]Storage[/]  [dim]Cached models and pre-download[/dim]")
            yield Static("[dim]Storage status appears here.[/dim]", id="storage-status")
            yield Static("")
            with Vertical(id="storage-controls"):
                yield Static("[bold]Backend filter[/bold]")
                yield OptionList(
                    Option("  All backends", id="filter-all"),
                    Option("  llama.cpp", id="filter-llamacpp"),
                    Option("  vLLM", id="filter-vllm"),
                    id="storage-backend-filter",
                )
                yield Static("")
                yield Static("[bold]Actions[/bold]")
                yield OptionList(
                    Option("  Refresh storage snapshot", id="refresh"),
                    Option("  Pre-download model", id="predownload"),
                    Option("  Delete selected model", id="delete"),
                    id="storage-action-list",
                )
            yield Static("")
            yield Static("[bold]Model inventory[/bold]")
            yield DataTable(id="storage-table")
            yield Static("[dim]Tip: select a row to prefill model fields.[/dim]", id="storage-hint")
            yield Static("")
            yield Input(placeholder="Model id (e.g. Qwen/Qwen3-4B-Thinking-2507-FP8)", id="storage-model-id")
            yield Input(
                placeholder="Backend for pre-download: llamacpp or vllm",
                id="storage-model-backend",
            )
            yield Input(placeholder="Quant pattern (llama.cpp only, optional)", id="storage-model-quant")
            yield Input(placeholder="Revision (optional)", id="storage-model-revision")
            yield Static("[dim]Press p to pre-download using these values.[/dim]")
        yield Footer()

    def on_mount(self) -> None:
        self._snapshot = StorageSnapshot(llamacpp_models=[], vllm_models=[])
        self._selected_filter: BackendType | None = self._initial_backend
        self._rows_by_key: dict[str, StoredModelInfo] = {}
        self._selected_model: StoredModelInfo | None = None
        self._was_suspended = False
        self._initial_focus_pending = True
        table = self.query_one("#storage-table", DataTable)
        table.cursor_type = "row"
        table.zebra_stripes = True
        table.add_columns("backend", "model", "revision", "quant", "files", "size")
        if self._initial_backend is not None:
            self.query_one("#storage-model-backend", Input).value = self._initial_backend.value
        backend_filter = self.query_one("#storage-backend-filter", OptionList)
        if self._selected_filter == BackendType.LLAMACPP:
            backend_filter.highlighted = 1
        elif self._selected_filter == BackendType.VLLM:
            backend_filter.highlighted = 2
        elif backend_filter.option_count > 0:
            backend_filter.highlighted = 0
        action_list = self.query_one("#storage-action-list", OptionList)
        if action_list.option_count > 0:
            action_list.highlighted = 0
        self.call_after_refresh(self._focus_first_visible_navigation_target)
        self._refresh_storage_snapshot()

    def on_screen_suspend(self, _: events.ScreenSuspend) -> None:
        self._was_suspended = True

    def on_screen_resume(self, _: events.ScreenResume) -> None:
        """Refresh storage when returning to this screen."""
        if self._was_suspended:
            self._was_suspended = False
            self._refresh_storage_snapshot()

    def on_option_list_option_selected(self, event: OptionList.OptionSelected) -> None:
        if event.option_list.id == "storage-backend-filter":
            if event.option.id == "filter-llamacpp":
                self._selected_filter = BackendType.LLAMACPP
            elif event.option.id == "filter-vllm":
                self._selected_filter = BackendType.VLLM
            else:
                self._selected_filter = None
            self._render_table()
            return

        if event.option_list.id == "storage-action-list":
            if event.option.id == "refresh":
                self.action_refresh_storage()
            elif event.option.id == "predownload":
                self.action_predownload_selected()
            elif event.option.id == "delete":
                self.action_delete_selected_model()

    def on_data_table_row_selected(self, event: DataTable.RowSelected) -> None:
        row_key = str(event.row_key.value) if getattr(event, "row_key", None) is not None else ""
        self._apply_row_selection(row_key)

    def on_data_table_row_highlighted(self, event: DataTable.RowHighlighted) -> None:
        row_key = str(event.row_key.value) if getattr(event, "row_key", None) is not None else ""
        self._apply_row_selection(row_key)

    def on_storage_loaded(self, message: StorageLoaded) -> None:
        self._snapshot = message.snapshot
        self._render_table()
        self.query_one("#storage-status", Static).update(
            "[green]Storage refreshed.[/green] Use selected row or type a model to pre-download."
        )
        should_refocus = self._initial_focus_pending
        self._initial_focus_pending = False
        focused = self.focused
        if should_refocus or not isinstance(focused, Widget) or not _is_focusable_for_arrow_navigation(focused):
            self.call_after_refresh(self._focus_first_visible_navigation_target)

    def on_storage_failed(self, message: StorageFailed) -> None:
        self.query_one("#storage-status", Static).update(
            f"[yellow]Storage refresh failed:[/yellow] {message.error}"
        )

    def _render_table(self) -> None:
        table = self.query_one("#storage-table", DataTable)
        table.clear(columns=False)
        self._rows_by_key = {}

        rows = self._snapshot.llamacpp_models + self._snapshot.vllm_models
        if self._selected_filter is not None:
            rows = [row for row in rows if row.backend == self._selected_filter]

        for index, row in enumerate(rows):
            row_key = f"{row.backend.value}:{index}:{row.model_id}"
            self._rows_by_key[row_key] = row
            table.add_row(
                row.backend.value,
                _model_label(row),
                row.revision or "-",
                row.quant or "-",
                str(row.file_count),
                _human_bytes(row.size_bytes),
                key=row_key,
            )

    def _apply_row_selection(self, row_key: str) -> None:
        selected = self._rows_by_key.get(row_key)
        if selected is None:
            return
        self._selected_model = selected
        self.query_one("#storage-model-id", Input).value = selected.model_id
        self.query_one("#storage-model-backend", Input).value = selected.backend.value
        self.query_one("#storage-model-quant", Input).value = selected.quant or ""
        self.query_one("#storage-model-revision", Input).value = selected.revision or ""

    def action_refresh_storage(self) -> None:
        self._refresh_storage_snapshot(force=True)

    def _refresh_storage_snapshot(self, force: bool = False) -> None:
        self.query_one("#storage-status", Static).update("[dim]Refreshing storage snapshot...[/dim]")
        self.app.begin_storage_refresh(self, force=force)  # type: ignore[attr-defined]

    def action_predownload_selected(self) -> None:
        model_id = self.query_one("#storage-model-id", Input).value.strip()
        backend_raw = self.query_one("#storage-model-backend", Input).value.strip().lower()
        quant = self.query_one("#storage-model-quant", Input).value.strip() or None
        revision = self.query_one("#storage-model-revision", Input).value.strip() or None
        if not model_id:
            self.app.notify("Model id is required for pre-download.", severity="error", timeout=5)
            return
        if backend_raw not in {"llamacpp", "vllm"}:
            self.app.notify("Backend must be either 'llamacpp' or 'vllm'.", severity="error", timeout=5)
            return
        backend = BackendType(backend_raw)
        if backend == BackendType.VLLM:
            quant = None
        self.app.begin_storage_predownload(  # type: ignore[attr-defined]
            backend=backend,
            model_id=model_id,
            quant=quant,
            revision=revision,
        )

    def action_delete_selected_model(self) -> None:
        if self._selected_model is None:
            self.app.notify("Select a model row first to delete.", severity="warning", timeout=5)
            return
        self.app.begin_storage_delete(self._selected_model)  # type: ignore[attr-defined]

    def action_pop_screen(self) -> None:
        self.app.pop_screen()

    def action_navigate_option_list_down(self) -> None:
        if move_focus_across_option_lists(
            self,
            ("storage-backend-filter", "storage-action-list"),
            direction=1,
            is_focusable=_is_focusable_for_arrow_navigation,
        ):
            return
        focused = self.focused
        if isinstance(focused, DataTable) and focused.row_count > 0 and focused.cursor_row < focused.row_count - 1:
            raise SkipAction()
        if move_focus_across_widgets(
            self,
            self.NAVIGATION_ORDER,
            direction=1,
            is_focusable=_is_focusable_for_arrow_navigation,
            fallback_to_edge_if_focus_missing=True,
        ):
            return
        raise SkipAction()

    def action_navigate_option_list_up(self) -> None:
        if move_focus_across_option_lists(
            self,
            ("storage-backend-filter", "storage-action-list"),
            direction=-1,
            is_focusable=_is_focusable_for_arrow_navigation,
        ):
            return
        focused = self.focused
        if isinstance(focused, DataTable) and focused.row_count > 0 and focused.cursor_row > 0:
            raise SkipAction()
        if move_focus_across_widgets(
            self,
            self.NAVIGATION_ORDER,
            direction=-1,
            is_focusable=_is_focusable_for_arrow_navigation,
            fallback_to_edge_if_focus_missing=True,
        ):
            return
        raise SkipAction()

    def _focus_first_visible_navigation_target(self) -> None:
        for widget_id in self.NAVIGATION_ORDER:
            widget = self.query_one(f"#{widget_id}", Widget)
            if not _is_focusable_for_arrow_navigation(widget):
                continue
            widget.focus()
            if isinstance(widget, OptionList) and widget.highlighted is None:
                widget.highlighted = first_enabled_option_index(widget)
            return



"""High-level workflow orchestrator that yields protocol events.

All long-running operations (deploy, warmup, logs, etc.) are implemented
as generators that yield protocol events. The TUI workers and headless
CLI both consume these generators.
"""

from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor
from dataclasses import replace
import json
import os
import queue
import re
import subprocess
import threading
import time
from typing import Any, Generator, List, Optional

from ..protocol.enums import BackendType, DeploymentState, OperationType
from ..protocol.events import (
    BaseEvent,
    ErrorEvent,
    LogEvent,
    OperationCompleteEvent,
    StateChangeEvent,
)
from ..protocol.models import DeploymentConfig, EndpointInfo
from ..protocol.models import StoredModelInfo, StorageSnapshot

from .backend import ModalBackend
from .config import ConfigStore
from .modal_auth import get_modal_auth_status
from .naming import legacy_app_name
from .naming import default_llamacpp_served_model_name
from .naming import random_function_slug
from .paths import MODAL_LLAMACPP_SCRIPT, MODAL_VLLM_SCRIPT

# Type alias for event generators
EventStream = Generator[BaseEvent, None, None]

_MODAL_GPU_WAIT_RE = re.compile(
    r"waiting to be scheduled on a (?P<worker>[A-Za-z0-9_:\-]+) worker",
    flags=re.IGNORECASE,
)
_MODAL_RELAX_RE = re.compile(r"Relaxing requirements \((?P<requirements>[^)]+)\)", flags=re.IGNORECASE)
_GGUF_QUANT_RE = re.compile(r"(Q\d(?:_[A-Z0-9]+)+|IQ\d+_[A-Z0-9_]+)", flags=re.IGNORECASE)
_SIZE_TOKEN_RE = re.compile(r"^\s*(?P<num>\d+(?:\.\d+)?)\s*(?P<unit>[A-Za-z]+)?\s*$")
_LLAMACPP_MIN_PLAUSIBLE_GGUF_BYTES = 1024 * 1024
_LLAMACPP_STORAGE_JSON_BEGIN = "LLM_LAUNCHPAD_STORAGE_JSON_BEGIN"
_LLAMACPP_STORAGE_JSON_END = "LLM_LAUNCHPAD_STORAGE_JSON_END"


def _is_historical_modal_app_state(state: str) -> bool:
    """Return True for terminal/inactive Modal app rows that commonly accumulate."""
    normalized = (state or "").strip().lower()
    return normalized in {"stopped", "stopping", "terminated", "archived"}


def _dedupe_launchpad_apps(rows: list[EndpointInfo]) -> list[EndpointInfo]:
    """Collapse repeated historical rows while preserving concurrent live apps.

    Modal app list often includes many stopped revisions for one logical app.
    Those should not crowd the UI, but two active rows with the same app name
    can represent distinct live jobs and must both remain visible.
    """
    active_keys = {
        (
            row.backend.value if row.backend else "",
            (row.instance_name or "").strip(),
            (row.name or "").strip(),
        )
        for row in rows
        if not _is_historical_modal_app_state(row.state)
    }
    kept_historical_keys: set[tuple[str, str, str]] = set()
    deduped: list[EndpointInfo] = []

    for row in rows:
        key = (
            row.backend.value if row.backend else "",
            (row.instance_name or "").strip(),
            (row.name or "").strip(),
        )
        if not _is_historical_modal_app_state(row.state):
            deduped.append(row)
            continue
        if key in active_keys or key in kept_historical_keys:
            continue
        kept_historical_keys.add(key)
        deduped.append(row)

    return deduped


def _modal_gpu_scheduling_hint(response_text: str) -> str | None:
    """Extract a concise queueing status from Modal's scheduling message."""
    text = (response_text or "").strip()
    if not text:
        return None
    if "waiting to be scheduled" not in text.lower():
        return None

    worker = ""
    requirements = ""
    worker_match = _MODAL_GPU_WAIT_RE.search(text)
    if worker_match:
        worker = worker_match.group("worker").strip()
    requirements_match = _MODAL_RELAX_RE.search(text)
    if requirements_match:
        requirements = requirements_match.group("requirements").strip()

    parts: list[str] = []
    if worker:
        parts.append(worker)
    if requirements:
        parts.append(requirements)

    if not parts:
        return "Waiting for GPU scheduling"
    return f"Waiting for GPU scheduling ({', '.join(parts)})"


def _probe_response_is_ready(backend: BackendType, status_code: int, body: str) -> tuple[bool, str | None]:
    """Validate readiness probe success beyond HTTP status for llama.cpp.

    Modal can return intermediary/fallback 200 responses while the underlying
    containerized server is still starting or has already crashed. For llama.cpp
    we require an OpenAI-compatible completions payload (`choices`) to treat the
    endpoint as ready.
    """
    if not (200 <= status_code < 300):
        return False, None
    if backend == BackendType.VLLM:
        return True, None

    text = (body or "").strip()
    if not text:
        return False, "empty response body"
    try:
        payload = json.loads(text)
    except json.JSONDecodeError:
        return False, "non-JSON response body"
    if not isinstance(payload, dict):
        return False, "unexpected JSON payload type"
    choices = payload.get("choices")
    if isinstance(choices, list):
        return True, None
    if isinstance(payload.get("error"), dict):
        return False, "error response"
    return False, "missing completions choices"


class Orchestrator:
    """Coordinate high-level deployment flows.

    Every public method is a generator yielding protocol events so callers
    can stream output in real-time (TUI workers, headless printers, etc.).
    """

    _LLAMACPP_STORAGE_VOLUME = "huggingface-cache"
    _VLLM_STORAGE_VOLUME = "huggingface-cache"

    def __init__(
        self,
        config_store: ConfigStore | None = None,
        backend: ModalBackend | None = None,
    ) -> None:
        self.config_store = config_store or ConfigStore()
        self.backend = backend or ModalBackend()

    # ------------------------------------------------------------------
    # Pre-flight
    # ------------------------------------------------------------------

    def preflight(self) -> tuple[bool, str, str]:
        """Check Modal CLI availability and auth.

        Returns ``(ok, username, error_message)``.
        """
        if not ModalBackend.is_cli_available():
            return False, "", "Modal CLI not found. Reinstall llm-launchpad, then run: modal setup"
        status = get_modal_auth_status()
        if not status.authenticated:
            return False, "", "Modal authentication missing. Run: modal setup"
        return True, status.profile or "", ""

    # ------------------------------------------------------------------
    # Deploy
    # ------------------------------------------------------------------

    def deploy(self, config: DeploymentConfig) -> EventStream:
        """Run a full deploy workflow (optional preload + deploy + warmup)."""
        if config.do_deploy and config.backend != BackendType.VLLM and not config.function_slug:
            config.function_slug = random_function_slug()
        if config.backend == BackendType.LLAMACPP and not (config.served_model_name or "").strip():
            config.served_model_name = default_llamacpp_served_model_name(
                config.repo_id,
                config.quant,
            )
        settings = self.config_store.load()
        env = ModalBackend.build_full_env(settings, config)

        if config.backend == BackendType.VLLM:
            yield from self._deploy_vllm(config, env)
        else:
            yield from self._deploy_llamacpp(config, env)

    def _deploy_vllm(
        self, config: DeploymentConfig, env: dict[str, str]
    ) -> EventStream:
        if config.do_deploy:
            cmd = ModalBackend.build_deploy_command(config.backend, app_name=config.app_name)
            yield StateChangeEvent(
                current=DeploymentState.DEPLOYING,
                operation=OperationType.DEPLOY,
                detail=" ".join(cmd),
            )
            yield LogEvent(line=f"Running: {' '.join(cmd)}")
            if env:
                yield LogEvent(line=f"  env: {', '.join(f'{k}={v}' for k, v in env.items())}")
            yield from ModalBackend.run_streaming(cmd, env=env)
        elif config.run_smoke:
            cmd = ["modal", "run", BackendType.VLLM.script]
            yield StateChangeEvent(
                current=DeploymentState.RUNNING,
                operation=OperationType.SMOKE_TEST,
                detail=" ".join(cmd),
            )
            yield LogEvent(line=f"Running: {' '.join(cmd)}")
            yield from ModalBackend.run_streaming(cmd, env=env)

    def _deploy_llamacpp(
        self, config: DeploymentConfig, env: dict[str, str]
    ) -> EventStream:
        def _run_step(
            cmd: list[str],
            *,
            state: DeploymentState,
            emit_env: bool,
        ) -> Generator[BaseEvent, None, tuple[bool, int, str]]:
            yield StateChangeEvent(
                current=state,
                operation=OperationType.DEPLOY,
                detail=" ".join(cmd),
            )
            yield LogEvent(line=f"Running: {' '.join(cmd)}")
            if emit_env and env:
                yield LogEvent(line=f"  env: {', '.join(f'{k}={v}' for k, v in env.items())}")

            saw_completion = False
            success = True
            exit_code = 0
            detail = ""
            for event in ModalBackend.run_streaming(cmd, env=env):
                if isinstance(event, OperationCompleteEvent):
                    saw_completion = True
                    success = event.success
                    exit_code = event.exit_code
                    detail = event.detail
                    continue
                if isinstance(event, ErrorEvent):
                    yield ErrorEvent(
                        message=event.message,
                        operation=OperationType.DEPLOY,
                        exit_code=event.exit_code,
                        recoverable=event.recoverable,
                    )
                    return False, event.exit_code or 1, event.message
                if isinstance(event, LogEvent):
                    yield LogEvent(
                        line=event.line,
                        stream=event.stream,
                        operation=OperationType.DEPLOY,
                    )
                    continue
                yield event

            if not saw_completion:
                return False, 1, "Command finished without completion event."
            return success, exit_code, detail

        if not config.do_deploy:
            cmd = ModalBackend.build_run_command(config)
            ok, exit_code, detail = yield from _run_step(
                cmd,
                state=DeploymentState.RUNNING,
                emit_env=True,
            )
            yield OperationCompleteEvent(
                operation=OperationType.DEPLOY,
                success=ok,
                exit_code=exit_code,
                detail=detail,
            )
            return

        # Split llama.cpp flow into:
        # 1) configure + optional preload (`modal run ...::main` without nested deploy)
        # 2) explicit `modal deploy`
        prep_config = replace(config, do_deploy=False)
        prep_cmd = ModalBackend.build_run_command(prep_config)
        prep_ok, prep_exit_code, prep_detail = yield from _run_step(
            prep_cmd,
            state=DeploymentState.DEPLOYING,
            emit_env=True,
        )
        if not prep_ok:
            yield OperationCompleteEvent(
                operation=OperationType.DEPLOY,
                success=False,
                exit_code=prep_exit_code,
                detail=prep_detail or "llama.cpp preparation step failed",
            )
            return

        deploy_cmd = ModalBackend.build_deploy_command(config.backend, app_name=config.app_name)
        deploy_ok, deploy_exit_code, deploy_detail = yield from _run_step(
            deploy_cmd,
            state=DeploymentState.DEPLOYING,
            emit_env=False,
        )
        yield OperationCompleteEvent(
            operation=OperationType.DEPLOY,
            success=deploy_ok,
            exit_code=deploy_exit_code,
            detail=deploy_detail,
        )

    # ------------------------------------------------------------------
    # Warmup
    # ------------------------------------------------------------------

    def warmup(
        self,
        backend: BackendType,
        server_url: str,
        timeout: int = 1800,
        tail_logs: bool = True,
        app_name: Optional[str] = None,
        served_model_name: Optional[str] = None,
    ) -> EventStream:
        """Probe endpoint readiness and optionally tail logs."""
        yield StateChangeEvent(
            current=DeploymentState.WARMING_UP,
            operation=OperationType.WARMUP,
            detail=f"Probing {server_url}",
        )

        is_vllm = backend == BackendType.VLLM
        probe_url = server_url.rstrip("/") + ("/health" if is_vllm else "/v1/completions")
        yield LogEvent(line=f"Probing readiness at: {probe_url}")

        # Start log tailing in background.
        # We use a dedicated reader thread + queue.Queue so that every
        # line the subprocess emits is captured immediately, regardless
        # of Python's internal TextIOWrapper read-ahead buffering.
        # (select() only checks the kernel pipe buffer, but readline()
        # can consume up to 2 KB into an internal buffer in a single
        # call — a mismatch that causes lines to be "stuck" until the
        # next kernel-buffer read.)
        target_app_name = app_name or legacy_app_name(backend)
        logs_proc: Optional[subprocess.Popen[str]] = None
        log_queue: Optional[queue.Queue[Optional[str]]] = None
        logs_retry_at = 0.0

        def _start_logs_tail() -> subprocess.Popen[str]:
            nonlocal log_queue
            follow = ModalBackend.logs_follow_args()
            # PYTHONUNBUFFERED forces the modal CLI (a Python program) to
            # flush each write immediately when stdout is a pipe.
            unbuf_env = {**os.environ, "PYTHONUNBUFFERED": "1"}
            proc = subprocess.Popen(
                ["modal", "app", "logs", *follow, target_app_name],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                text=True,
                bufsize=1,
                env=unbuf_env,
            )
            q: queue.Queue[Optional[str]] = queue.Queue()
            log_queue = q

            def _reader() -> None:
                try:
                    assert proc.stdout is not None
                    for raw_line in proc.stdout:
                        q.put(raw_line.rstrip("\n"))
                except Exception:
                    pass
                finally:
                    q.put(None)  # sentinel: reader finished

            threading.Thread(target=_reader, daemon=True).start()
            return proc

        if tail_logs:
            try:
                logs_proc = _start_logs_tail()
            except Exception as exc:
                yield LogEvent(line=f"Warning: failed to start log tailing: {exc}")
                logs_retry_at = time.time() + 5.0

        try:
            import requests
        except ImportError:
            yield ErrorEvent(
                message="'requests' is required. Install with: pip install requests",
                operation=OperationType.WARMUP,
                recoverable=False,
            )
            yield OperationCompleteEvent(
                operation=OperationType.WARMUP,
                success=False,
                exit_code=1,
            )
            return

        llama_probe_model = (served_model_name or "").strip() or "default"
        payload = {
            "model": llama_probe_model,
            "prompt": "ping",
            "max_tokens": 1,
            "temperature": 0,
        }
        headers = {"Content-Type": "application/json"}
        start = time.time()
        backoff = 2.0
        max_backoff = 30.0
        last_err: Optional[str] = None
        last_scheduling_hint: Optional[str] = None
        # Track displayed log lines so the historical-fetch fallback can
        # avoid duplicating output already shown by the live stream.
        seen_log_lines: set[str] = set()

        def _drain_queue() -> Generator[LogEvent, None, None]:
            """Yield log events from the reader-thread queue (non-blocking)."""
            nonlocal logs_proc, log_queue, logs_retry_at
            if log_queue is None:
                return
            while True:
                try:
                    line = log_queue.get_nowait()
                except queue.Empty:
                    break
                if line is None:
                    # Reader thread finished — subprocess exited.
                    logs_proc = None
                    log_queue = None
                    logs_retry_at = time.time() + 5.0
                    break
                seen_log_lines.add(line)
                yield LogEvent(line=line, operation=OperationType.WARMUP)

        while True:
            # Exit promptly if the app is shutting down.
            if ModalBackend.is_shutting_down():
                if logs_proc:
                    try:
                        logs_proc.terminate()
                    except Exception:
                        pass
                return

            if tail_logs and logs_proc is None and time.time() >= logs_retry_at:
                # ── Historical-fetch fallback ──
                # The live stream (`modal app logs`) may not deliver the
                # final output of a crashing container before the stream
                # closes.  The Modal dashboard shows those logs because it
                # reads *persisted* logs.  Re-running `modal app logs`
                # after a short delay retrieves the same persisted data,
                # letting us display crash tracebacks the live stream
                # missed.
                yield from self._fetch_historical_logs(
                    target_app_name, seen_log_lines
                )
                # Re-attach live stream for further output.
                try:
                    logs_proc = _start_logs_tail()
                except Exception as exc:
                    yield LogEvent(line=f"Warning: failed to start log tailing: {exc}")
                    logs_retry_at = time.time() + 5.0

            # Drain log lines delivered by the reader thread.
            yield from _drain_queue()

            elapsed = time.time() - start
            if elapsed > timeout:
                if logs_proc:
                    try:
                        logs_proc.terminate()
                    except Exception:
                        pass
                yield ErrorEvent(
                    message=f"Timed out after {timeout}s. Last error: {last_err}",
                    operation=OperationType.WARMUP,
                )
                yield OperationCompleteEvent(
                    operation=OperationType.WARMUP, success=False, exit_code=1
                )
                return

            try:
                import json as _json

                if is_vllm:
                    resp = requests.get(probe_url, timeout=10)
                else:
                    resp = requests.post(
                        probe_url, headers=headers, data=_json.dumps(payload), timeout=10
                    )
                body = resp.text or ""
                ready_ok, readiness_reason = _probe_response_is_ready(
                    backend, resp.status_code, body
                )
                if ready_ok:
                    if logs_proc:
                        try:
                            logs_proc.terminate()
                        except Exception:
                            pass
                    yield LogEvent(line="Server is ready!")
                    curl_cmd = ModalBackend.test_curl_command(
                        backend,
                        server_url,
                        served_model_name=served_model_name,
                    )
                    yield LogEvent(line=f"Test command:\n{curl_cmd}")
                    yield StateChangeEvent(
                        current=DeploymentState.HEALTHY, operation=OperationType.WARMUP
                    )
                    yield OperationCompleteEvent(
                        operation=OperationType.WARMUP, success=True, data={"url": server_url}
                    )
                    return
                if 200 <= resp.status_code < 300 and readiness_reason:
                    last_err = f"HTTP {resp.status_code} (not ready): {readiness_reason}; body={body[:200]}"
                else:
                    last_err = f"HTTP {resp.status_code}: {body[:200]}"
                scheduling_hint = _modal_gpu_scheduling_hint(body)
                if scheduling_hint:
                    if scheduling_hint != last_scheduling_hint:
                        yield StateChangeEvent(
                            current=DeploymentState.QUEUED,
                            operation=OperationType.WARMUP,
                            detail=scheduling_hint,
                        )
                        yield LogEvent(line=scheduling_hint, operation=OperationType.WARMUP)
                        last_scheduling_hint = scheduling_hint
                elif last_scheduling_hint is not None:
                    yield StateChangeEvent(
                        current=DeploymentState.WARMING_UP,
                        operation=OperationType.WARMUP,
                        detail="GPU allocated; continuing readiness probe",
                    )
                    yield LogEvent(
                        line="GPU allocated; continuing readiness probe.",
                        operation=OperationType.WARMUP,
                    )
                    last_scheduling_hint = None
            except Exception as exc:
                last_err = str(exc)

            # Sleep in small increments, draining the log queue between
            # each chunk so that lines appear in the TUI promptly.
            sleep_end = time.time() + backoff
            while time.time() < sleep_end:
                if ModalBackend.is_shutting_down():
                    break
                chunk = min(0.5, sleep_end - time.time())
                if chunk > 0:
                    ModalBackend._shutdown_event.wait(timeout=chunk)
                yield from _drain_queue()
            backoff = min(max_backoff, backoff * 1.5)

    # ------------------------------------------------------------------
    # Historical log fetch (fallback for crash output)
    # ------------------------------------------------------------------

    @staticmethod
    def _fetch_historical_logs(
        app_name: str,
        seen: set[str],
    ) -> EventStream:
        """Re-run ``modal app logs`` to capture persisted crash output.

        The live log stream may not deliver the tail end of a crashing
        container's output.  Modal persists those logs server-side and
        the dashboard reads them.  By re-running the CLI command after a
        short delay we retrieve the same data, then yield only the lines
        the live stream missed (tracked via *seen*).
        """
        try:
            unbuf_env = {**os.environ, "PYTHONUNBUFFERED": "1"}
            hist = subprocess.run(
                ["modal", "app", "logs", app_name],
                capture_output=True,
                text=True,
                timeout=15,
                env=unbuf_env,
            )
            hist_output = hist.stdout or ""
        except subprocess.TimeoutExpired as te:
            # Use whatever output was captured before the timeout.
            hist_output = te.stdout if isinstance(te.stdout, str) else ""
        except Exception:
            hist_output = ""

        if not hist_output:
            return

        new_lines: list[str] = []
        for raw_line in hist_output.splitlines():
            line = raw_line.rstrip()
            if line and line not in seen:
                new_lines.append(line)
                seen.add(line)

        if new_lines:
            yield LogEvent(
                line="── Fetched historical container logs ──",
                operation=OperationType.WARMUP,
            )
            for line in new_lines:
                yield LogEvent(line=line, operation=OperationType.WARMUP)

    # ------------------------------------------------------------------
    # Logs
    # ------------------------------------------------------------------

    def tail_logs(
        self,
        backend: BackendType,
        follow: bool = True,
        app_name: Optional[str] = None,
        app_id: Optional[str] = None,
    ) -> EventStream:
        """Tail Modal logs for the given backend."""
        target_app_name = (app_id or app_name or legacy_app_name(backend)).strip()
        cmd: List[str] = ["modal", "app", "logs"]
        if follow:
            cmd.extend(ModalBackend.logs_follow_args())
        cmd.append(target_app_name)

        yield StateChangeEvent(
            current=DeploymentState.RUNNING,
            operation=OperationType.LOGS,
            detail=" ".join(cmd),
        )
        yield from ModalBackend.run_streaming(cmd)

    # ------------------------------------------------------------------
    # Status
    # ------------------------------------------------------------------

    def check_status(
        self,
        backend: BackendType,
        server_url: str,
        timeout: int = 60,
        served_model_name: Optional[str] = None,
    ) -> EventStream:
        """Probe endpoint health with backoff."""
        is_vllm = backend == BackendType.VLLM
        probe_url = server_url.rstrip("/") + ("/health" if is_vllm else "/v1/completions")

        yield StateChangeEvent(
            current=DeploymentState.RUNNING,
            operation=OperationType.STATUS,
            detail=f"Checking {probe_url}",
        )
        yield LogEvent(line=f"Checking endpoint status at: {probe_url}")

        try:
            import requests
        except ImportError:
            yield ErrorEvent(
                message="'requests' required", operation=OperationType.STATUS, recoverable=False
            )
            yield OperationCompleteEvent(
                operation=OperationType.STATUS,
                success=False,
                exit_code=1,
            )
            return

        import json as _json

        payload = {
            "model": (served_model_name or "").strip() or "default",
            "prompt": "ping",
            "max_tokens": 1,
            "temperature": 0,
        }
        headers = {"Content-Type": "application/json"}
        start = time.time()
        backoff = 2.0
        max_backoff = 15.0
        last_err: Optional[str] = None
        last_scheduling_hint: Optional[str] = None

        while True:
            if ModalBackend.is_shutting_down():
                return

            if time.time() - start > timeout:
                yield ErrorEvent(
                    message=f"Unhealthy (timed out). Last: {last_err}",
                    operation=OperationType.STATUS,
                )
                yield OperationCompleteEvent(
                    operation=OperationType.STATUS, success=False, exit_code=1
                )
                return

            try:
                if is_vllm:
                    resp = requests.get(probe_url, timeout=10)
                else:
                    resp = requests.post(
                        probe_url, headers=headers, data=_json.dumps(payload), timeout=10
                    )
                body = resp.text or ""
                ready_ok, readiness_reason = _probe_response_is_ready(
                    backend, resp.status_code, body
                )
                if ready_ok:
                    curl_cmd = ModalBackend.test_curl_command(
                        backend,
                        server_url,
                        served_model_name=served_model_name,
                    )
                    yield LogEvent(
                        line=f"Status: healthy (backend={backend.value}, url={server_url})"
                    )
                    yield LogEvent(line=f"Test command:\n{curl_cmd}")
                    yield StateChangeEvent(
                        current=DeploymentState.HEALTHY, operation=OperationType.STATUS
                    )
                    yield OperationCompleteEvent(
                        operation=OperationType.STATUS, success=True
                    )
                    return
                if 200 <= resp.status_code < 300 and readiness_reason:
                    last_err = f"HTTP {resp.status_code} (not ready): {readiness_reason}; body={body[:200]}"
                else:
                    last_err = f"HTTP {resp.status_code}: {body[:200]}"
                scheduling_hint = _modal_gpu_scheduling_hint(body)
                if scheduling_hint:
                    if scheduling_hint != last_scheduling_hint:
                        yield StateChangeEvent(
                            current=DeploymentState.QUEUED,
                            operation=OperationType.STATUS,
                            detail=scheduling_hint,
                        )
                        yield LogEvent(line=scheduling_hint, operation=OperationType.STATUS)
                        last_scheduling_hint = scheduling_hint
                elif last_scheduling_hint is not None:
                    yield StateChangeEvent(
                        current=DeploymentState.RUNNING,
                        operation=OperationType.STATUS,
                        detail="GPU allocated; continuing health check",
                    )
                    yield LogEvent(
                        line="GPU allocated; continuing health check.",
                        operation=OperationType.STATUS,
                    )
                    last_scheduling_hint = None
            except Exception as exc:
                last_err = str(exc)

            # Interruptible sleep: wakes immediately on shutdown.
            if ModalBackend._shutdown_event.wait(timeout=backoff):
                return
            backoff = min(max_backoff, backoff * 1.5)

    # ------------------------------------------------------------------
    # List
    # ------------------------------------------------------------------

    def list_deployments(self) -> EventStream:
        """List launchpad-related Modal apps."""
        yield StateChangeEvent(
            current=DeploymentState.RUNNING, operation=OperationType.LIST
        )

        apps = ModalBackend.list_apps()
        if apps is not None:
            launchpad = [a for a in apps if a.backend is not None]
            visible_launchpad = _dedupe_launchpad_apps(launchpad)
            if not launchpad:
                yield LogEvent(line="No launchpad deployments found.")
            else:
                yield LogEvent(line="Launchpad deployments:")
                for info in visible_launchpad:
                    suffix = f" ({info.app_id})" if info.app_id else ""
                    bk = info.backend.value if info.backend else "unknown"
                    inst = info.instance_name or "-"
                    yield LogEvent(
                        line=f"  backend={bk}  instance={inst}  app={info.name}  state={info.state}{suffix}"
                    )
                hidden_count = len(launchpad) - len(visible_launchpad)
                if hidden_count > 0:
                    yield LogEvent(
                        line=f"  [hidden {hidden_count} historical duplicate app row{'s' if hidden_count != 1 else ''}]"
                    )
            yield OperationCompleteEvent(
                operation=OperationType.LIST, success=True, data=visible_launchpad
            )
            return

        # Fallback to raw text
        raw = ModalBackend.list_apps_raw()
        if raw:
            names = {legacy_app_name(bt) for bt in BackendType}
            lines = raw.splitlines()
            matches = [
                line
                for line in lines
                if any(name in line for name in names) or "vllm-" in line or "llamacpp-" in line
            ]
            if matches:
                yield LogEvent(line="Launchpad deployments:")
                for line in matches:
                    yield LogEvent(line=f"  {line.strip()}")
            else:
                yield LogEvent(line="No launchpad deployments found.")
            yield OperationCompleteEvent(operation=OperationType.LIST, success=True)
        else:
            yield ErrorEvent(
                message="Failed to query Modal app list.",
                operation=OperationType.LIST,
            )
            yield OperationCompleteEvent(
                operation=OperationType.LIST, success=False, exit_code=1
            )

    # ------------------------------------------------------------------
    # Storage
    # ------------------------------------------------------------------

    def list_storage(self) -> EventStream:
        """List cached model artifacts for both backends."""
        yield StateChangeEvent(
            current=DeploymentState.RUNNING,
            operation=OperationType.STORAGE_LIST,
            detail="Scanning Modal volumes for cached models",
        )
        scan_started = time.perf_counter()
        try:
            with ThreadPoolExecutor(max_workers=2) as executor:
                llamacpp_future = executor.submit(self._list_llamacpp_models)
                vllm_future = executor.submit(self._list_vllm_models)
                snapshot = StorageSnapshot(
                    llamacpp_models=llamacpp_future.result(),
                    vllm_models=vllm_future.result(),
                )
        except Exception as exc:
            yield ErrorEvent(
                message=f"Failed to list storage: {exc}",
                operation=OperationType.STORAGE_LIST,
                recoverable=True,
            )
            yield OperationCompleteEvent(
                operation=OperationType.STORAGE_LIST,
                success=False,
                exit_code=1,
                detail=str(exc),
            )
            return

        scan_elapsed_ms = int((time.perf_counter() - scan_started) * 1000)
        yield LogEvent(
            line=(
                f"Storage snapshot: llama.cpp={len(snapshot.llamacpp_models)} models, "
                f"vLLM={len(snapshot.vllm_models)} models, "
                f"total={snapshot.total_models}, bytes={snapshot.total_size_bytes}, "
                f"scan_ms={scan_elapsed_ms}"
            ),
            operation=OperationType.STORAGE_LIST,
        )
        yield OperationCompleteEvent(
            operation=OperationType.STORAGE_LIST,
            success=True,
            data=snapshot,
        )

    def predownload_model(
        self,
        backend: BackendType,
        model_id: str,
        quant: Optional[str] = None,
        revision: Optional[str] = None,
    ) -> EventStream:
        """Pre-download a model for a backend without deploying."""
        if backend == BackendType.LLAMACPP:
            script = MODAL_LLAMACPP_SCRIPT
            entrypoint = "predownload_model"
            args = ["--repo-id", model_id]
            if quant:
                args.extend(["--quant", quant])
            if revision:
                args.extend(["--revision", revision])
        else:
            script = MODAL_VLLM_SCRIPT
            entrypoint = "predownload_model"
            args = ["--repo-id", model_id]
            if revision:
                args.extend(["--revision", revision])

        cmd = ModalBackend.build_modal_entrypoint_command(script, entrypoint, args)
        yield StateChangeEvent(
            current=DeploymentState.RUNNING,
            operation=OperationType.STORAGE_PREDOWNLOAD,
            detail=" ".join(cmd),
        )
        yield LogEvent(
            line=f"Running: {' '.join(cmd)}",
            operation=OperationType.STORAGE_PREDOWNLOAD,
        )

        for event in ModalBackend.run_modal_script_entrypoint(script, entrypoint, args=args):
            if isinstance(event, OperationCompleteEvent):
                yield OperationCompleteEvent(
                    operation=OperationType.STORAGE_PREDOWNLOAD,
                    success=event.success,
                    exit_code=event.exit_code,
                    detail=event.detail,
                    data=event.data,
                )
            elif isinstance(event, ErrorEvent):
                yield ErrorEvent(
                    message=event.message,
                    operation=OperationType.STORAGE_PREDOWNLOAD,
                    exit_code=event.exit_code,
                    recoverable=event.recoverable,
                )
            elif isinstance(event, LogEvent):
                yield LogEvent(
                    line=event.line,
                    stream=event.stream,
                    operation=OperationType.STORAGE_PREDOWNLOAD,
                )
            else:
                yield event

    def delete_stored_model(self, model: StoredModelInfo) -> EventStream:
        """Delete cached storage artifacts for a selected model."""
        targets = self._storage_delete_targets(model)
        if not targets:
            yield ErrorEvent(
                message=f"No removable paths found for {model.model_id}",
                operation=OperationType.STORAGE_DELETE,
                recoverable=True,
            )
            yield OperationCompleteEvent(
                operation=OperationType.STORAGE_DELETE,
                success=False,
                exit_code=1,
                detail="No paths resolved for deletion.",
            )
            return

        yield StateChangeEvent(
            current=DeploymentState.RUNNING,
            operation=OperationType.STORAGE_DELETE,
            detail=f"Deleting {model.model_id}",
        )
        for volume_name, remote_path, recursive in targets:
            yield LogEvent(
                line=f"Deleting {remote_path} from {volume_name}",
                operation=OperationType.STORAGE_DELETE,
            )
            for event in ModalBackend.run_volume_remove(volume_name, remote_path, recursive=recursive):
                if isinstance(event, OperationCompleteEvent):
                    if not event.success:
                        yield OperationCompleteEvent(
                            operation=OperationType.STORAGE_DELETE,
                            success=False,
                            exit_code=event.exit_code,
                            detail=event.detail or f"Failed removing {remote_path}",
                        )
                        return
                    continue
                if isinstance(event, ErrorEvent):
                    yield ErrorEvent(
                        message=event.message,
                        operation=OperationType.STORAGE_DELETE,
                        exit_code=event.exit_code,
                        recoverable=event.recoverable,
                    )
                    yield OperationCompleteEvent(
                        operation=OperationType.STORAGE_DELETE,
                        success=False,
                        exit_code=event.exit_code or 1,
                        detail=event.message,
                    )
                    return
                if isinstance(event, LogEvent):
                    yield LogEvent(
                        line=event.line,
                        stream=event.stream,
                        operation=OperationType.STORAGE_DELETE,
                    )
                else:
                    yield event

        yield OperationCompleteEvent(
            operation=OperationType.STORAGE_DELETE,
            success=True,
            detail=f"Deleted storage for {model.model_id}",
        )

    def _list_llamacpp_models(self) -> list[StoredModelInfo]:
        backend_rows = self._list_llamacpp_models_via_backend()
        if backend_rows is not None:
            return backend_rows
        return self._list_llamacpp_models_from_volume()

    def _list_llamacpp_models_via_backend(self) -> list[StoredModelInfo] | None:
        captured = ModalBackend.run_modal_script_entrypoint_capture(
            MODAL_LLAMACPP_SCRIPT,
            "list_downloaded_models_json",
        )
        if not captured:
            return None
        returncode, stdout, _stderr = captured
        if returncode != 0 or not stdout.strip():
            return None

        lines = stdout.splitlines()
        try:
            start = lines.index(_LLAMACPP_STORAGE_JSON_BEGIN)
            end = lines.index(_LLAMACPP_STORAGE_JSON_END, start + 1)
        except ValueError:
            return None
        if end <= start + 1:
            return []

        payload_text = "\n".join(lines[start + 1 : end]).strip()
        if not payload_text:
            return []
        try:
            payload = json.loads(payload_text)
        except Exception:
            return None
        if not isinstance(payload, list):
            return None

        rows: list[StoredModelInfo] = []
        for item in payload:
            if not isinstance(item, dict):
                continue
            model_id = str(item.get("model_id", "")).strip()
            if not model_id:
                continue
            revision_raw = item.get("revision")
            quant_raw = item.get("quant")
            source_volume_raw = item.get("source_volume")
            paths_raw = item.get("paths")
            rows.append(
                StoredModelInfo(
                    backend=BackendType.LLAMACPP,
                    model_id=model_id,
                    revision=str(revision_raw).strip() if revision_raw not in (None, "") else None,
                    quant=str(quant_raw).strip() if quant_raw not in (None, "") else None,
                    size_bytes=max(0, int(item.get("size_bytes", 0) or 0)),
                    file_count=max(0, int(item.get("file_count", 0) or 0)),
                    source_volume=(
                        str(source_volume_raw).strip()
                        if source_volume_raw not in (None, "")
                        else self._LLAMACPP_STORAGE_VOLUME
                    ),
                    paths=[str(p) for p in paths_raw if isinstance(p, str)]
                    if isinstance(paths_raw, list)
                    else [],
                    incomplete=False,
                )
            )
        return sorted(rows, key=lambda row: (row.model_id, row.revision or ""))

    def _list_llamacpp_models_from_volume(self) -> list[StoredModelInfo]:
        grouped: dict[tuple[str, Optional[str]], StoredModelInfo] = {}
        blob_files: dict[str, set[str]] = {}
        incomplete_blob_files: dict[str, set[str]] = {}
        snapshot_gguf_sizes: dict[str, list[int]] = {}
        hub_files = self._walk_volume_files(self._LLAMACPP_STORAGE_VOLUME, "/hub")
        for row in hub_files:
            path = str(row.get("path", "")).strip()
            size = int(row.get("size_bytes", 0) or 0)
            rel = path.removeprefix("/hub/").strip("/")
            if not rel.startswith("models--"):
                continue
            if not row.get("is_file", False):
                continue
            parts = rel.split("/")
            if len(parts) < 2:
                continue
            model_dir = parts[0]
            encoded = model_dir[len("models--") :]
            model_id = encoded.replace("--", "/")
            lower_path = path.lower()
            if "/blobs/" in lower_path:
                blob_name = path.rsplit("/", 1)[-1].strip()
                if blob_name.endswith(".incomplete"):
                    final_blob_name = blob_name[: -len(".incomplete")]
                    if final_blob_name:
                        incomplete_blob_files.setdefault(model_id, set()).add(final_blob_name)
                elif blob_name:
                    blob_files.setdefault(model_id, set()).add(blob_name)
            if not lower_path.endswith(".gguf"):
                continue
            snapshot_gguf_sizes.setdefault(model_id, []).append(size)
            key = (model_id, None)
            entry = grouped.get(key)
            if entry is None:
                entry = StoredModelInfo(
                    backend=BackendType.LLAMACPP,
                    model_id=model_id,
                    revision=None,
                    quant=None,
                    size_bytes=0,
                    file_count=0,
                    source_volume=self._LLAMACPP_STORAGE_VOLUME,
                    paths=[],
                )
                grouped[key] = entry
            entry.size_bytes += size
            entry.file_count += 1
            if entry.paths is not None and path not in entry.paths:
                entry.paths.append(path)
            if entry.quant is None:
                quant_match = _GGUF_QUANT_RE.search(path.upper())
                if quant_match:
                    entry.quant = quant_match.group(1).upper()

        for (model_id, _revision), entry in grouped.items():
            known_blobs = blob_files.get(model_id, set())
            pending_blobs = incomplete_blob_files.get(model_id, set())
            unresolved_incomplete = any(blob_name not in known_blobs for blob_name in pending_blobs)
            gguf_sizes = snapshot_gguf_sizes.get(model_id, [])
            tiny_snapshot_without_blob = (
                bool(gguf_sizes)
                and all(size < _LLAMACPP_MIN_PLAUSIBLE_GGUF_BYTES for size in gguf_sizes)
                and not known_blobs
            )
            if unresolved_incomplete or tiny_snapshot_without_blob:
                entry.incomplete = True

        return sorted(grouped.values(), key=lambda row: (row.model_id, row.revision or ""))

    def _list_vllm_models(self) -> list[StoredModelInfo]:
        files = self._walk_volume_files(self._VLLM_STORAGE_VOLUME, "/hub")
        grouped: dict[str, StoredModelInfo] = {}
        blob_files: dict[str, set[str]] = {}
        incomplete_blob_files: dict[str, set[str]] = {}
        snapshot_file_counts: dict[str, int] = {}
        ref_file_counts: dict[str, int] = {}
        gguf_models: set[str] = set()
        for row in files:
            path = str(row.get("path", "")).strip()
            size = int(row.get("size_bytes", 0) or 0)
            rel = path.removeprefix("/hub/").strip("/")
            if not rel.startswith("models--"):
                continue
            parts = rel.split("/")
            if not parts:
                continue
            model_dir = parts[0]
            encoded = model_dir[len("models--") :]
            model_id = encoded.replace("--", "/")
            entry = grouped.get(model_id)
            if entry is None:
                entry = StoredModelInfo(
                    backend=BackendType.VLLM,
                    model_id=model_id,
                    revision=None,
                    quant=None,
                    size_bytes=0,
                    file_count=0,
                    source_volume=self._VLLM_STORAGE_VOLUME,
                    paths=[],
                )
                grouped[model_id] = entry
            if row.get("is_file", False):
                entry.size_bytes += size
                entry.file_count += 1
                lower_path = path.lower()
                if "/snapshots/" in lower_path:
                    snapshot_file_counts[model_id] = snapshot_file_counts.get(model_id, 0) + 1
                if "/refs/" in lower_path:
                    ref_file_counts[model_id] = ref_file_counts.get(model_id, 0) + 1
                if "/blobs/" in lower_path:
                    blob_name = path.rsplit("/", 1)[-1].strip()
                    if blob_name.endswith(".incomplete"):
                        final_blob_name = blob_name[: -len(".incomplete")]
                        if final_blob_name:
                            incomplete_blob_files.setdefault(model_id, set()).add(final_blob_name)
                    elif blob_name:
                        blob_files.setdefault(model_id, set()).add(blob_name)
                if lower_path.endswith(".gguf"):
                    gguf_models.add(model_id)
                if entry.paths is not None and path not in entry.paths:
                    entry.paths.append(path)

        for model_id in list(grouped.keys()):
            if model_id in gguf_models:
                del grouped[model_id]

        if not grouped:
            return []

        for model_id, entry in grouped.items():
            known_blobs = blob_files.get(model_id, set())
            pending_blobs = incomplete_blob_files.get(model_id, set())
            snapshot_count = snapshot_file_counts.get(model_id, 0)
            ref_count = ref_file_counts.get(model_id, 0)
            # Align with huggingface_hub cache semantics:
            # incomplete blobs are resumable partial downloads, but once the
            # final blob exists, download logic treats that file as complete.
            # Keep stale orphan .incomplete blobs from old attempts from marking
            # a model incomplete when a usable snapshot already exists.
            unresolved_incomplete = any(blob_name not in known_blobs for blob_name in pending_blobs)
            if unresolved_incomplete and snapshot_count == 0:
                entry.incomplete = True
            # Refs without snapshots indicate metadata without an accessible
            # snapshot payload for this repo.
            if ref_count > 0 and snapshot_count == 0:
                entry.incomplete = True
        return sorted(grouped.values(), key=lambda row: row.model_id)

    def _walk_volume_files(
        self,
        volume_name: str,
        root: str,
        recursive: bool = True,
    ) -> list[dict[str, Any]]:
        pending = [root]
        files: list[dict[str, Any]] = []
        visited: set[str] = set()
        while pending:
            current = pending.pop()
            if current in visited:
                continue
            visited.add(current)
            rows = ModalBackend.list_volume(volume_name, current) or []
            for row in rows:
                path = self._volume_entry_path(current, row)
                if not path:
                    continue
                is_dir = self._volume_entry_is_dir(row, path)
                if is_dir:
                    if path == current or not recursive:
                        continue
                    pending.append(path)
                else:
                    files.append(
                        {
                            "path": path,
                            "size_bytes": Orchestrator._parse_size_bytes(
                                Orchestrator._entry_value(
                                    row, "size", "size_bytes", "bytes", "filesize"
                                )
                            ),
                            "is_file": True,
                        }
                    )
        return files

    def _storage_delete_targets(
        self, model: StoredModelInfo
    ) -> list[tuple[str, str, bool]]:
        targets: list[tuple[str, str, bool]] = []
        if model.backend == BackendType.LLAMACPP:
            encoded = model.model_id.replace("/", "--")
            targets.append(
                (
                    model.source_volume or self._LLAMACPP_STORAGE_VOLUME,
                    f"/hub/models--{encoded}",
                    True,
                )
            )
            return targets

        if model.backend == BackendType.VLLM:
            encoded = model.model_id.replace("/", "--")
            targets.append(
                (
                    model.source_volume or self._VLLM_STORAGE_VOLUME,
                    f"/hub/models--{encoded}",
                    True,
                )
            )
        return targets

    @staticmethod
    def _entry_value(row: dict[str, Any], *keys: str) -> Any:
        """Case-insensitive lookup for JSON fields from Modal CLI."""
        lowered = {str(key).strip().lower(): value for key, value in row.items()}
        for key in keys:
            needle = key.strip().lower()
            if needle in lowered:
                return lowered[needle]
        return None

    @staticmethod
    def _volume_entry_path(parent: str, row: dict[str, Any]) -> str:
        for key in ("path", "name", "file", "filename", "entry"):
            value = Orchestrator._entry_value(row, key)
            if isinstance(value, str) and value.strip():
                candidate = value.strip()
                if candidate.startswith("/"):
                    return candidate
                parent_clean = parent.strip("/")
                if parent_clean and candidate.startswith(f"{parent_clean}/"):
                    return f"/{candidate.lstrip('/')}"
                if parent == "/" and candidate:
                    return f"/{candidate.lstrip('/')}"
                if parent.endswith("/"):
                    return f"{parent}{candidate}"
                return f"{parent}/{candidate}"
        return ""

    @staticmethod
    def _volume_entry_is_dir(row: dict[str, Any], path: str) -> bool:
        is_dir = Orchestrator._entry_value(row, "is_dir", "isdir", "directory")
        if isinstance(is_dir, bool):
            return is_dir
        entry_type = str(Orchestrator._entry_value(row, "type", "kind") or "").strip().lower()
        if entry_type in {"dir", "directory", "folder"}:
            return True
        size_hint = Orchestrator._entry_value(row, "size", "size_bytes", "bytes")
        if size_hint in (None, "", 0, "0"):
            # If no file-size signal is present, treat slash-terminated entries as directories.
            if path.endswith("/"):
                return True
        return path.endswith("/")

    @staticmethod
    def _parse_size_bytes(value: Any) -> int:
        if value is None:
            return 0
        if isinstance(value, (int, float)):
            return int(value)
        text = str(value).strip()
        if not text:
            return 0
        clean = text.replace(",", "")
        if clean.isdigit():
            return int(clean)
        match = _SIZE_TOKEN_RE.match(clean)
        if not match:
            return 0
        number = float(match.group("num"))
        unit = (match.group("unit") or "B").upper()
        factors = {
            "B": 1,
            "KB": 1024,
            "MB": 1024**2,
            "GB": 1024**3,
            "TB": 1024**4,
            "KIB": 1024,
            "MIB": 1024**2,
            "GIB": 1024**3,
            "TIB": 1024**4,
        }
        factor = factors.get(unit, 1)
        return int(number * factor)

    # ------------------------------------------------------------------
    # Stop
    # ------------------------------------------------------------------

    def stop_app(
        self,
        backend: BackendType,
        app_name: Optional[str] = None,
        app_id: Optional[str] = None,
    ) -> EventStream:
        """Stop a deployed app."""
        target_app_name = (app_id or app_name or legacy_app_name(backend)).strip()
        cmd = ["modal", "app", "stop", target_app_name]
        yield StateChangeEvent(
            current=DeploymentState.RUNNING,
            operation=OperationType.STOP,
            detail=f"Stopping {target_app_name}",
        )
        yield LogEvent(line=f"Stopping app: {target_app_name}")
        yield from ModalBackend.run_streaming(cmd)
